from __future__ import absolute_import

import os
import textwrap

import matplotlib as mpl
import matplotlib.pyplot as plt

from kivy.properties import ObjectProperty, BooleanProperty
from kivy.uix.modalview import ModalView
from kivy.core.window import Window

from es_gui.resources.widgets.common import ResultsViewer


class BtmResultsViewer(ResultsViewer):
    """The screen for displaying plots inside the application or exporting results."""
    current_fig = ObjectProperty()
    current_ax = ObjectProperty()

    def __init__(self, **kwargs):
        super(BtmResultsViewer, self).__init__(**kwargs)

        self.dfs = {}

        self.time_selector.start_time.bind(on_text_validate=self.draw_figure)
        self.time_selector.end_time.bind(on_text_validate=self.draw_figure)

    def on_pre_enter(self):
        #Window.bind(on_key_down=self._on_keyboard_down)

        self._update_toolbar()

        # self.rv.data = self.manager.get_screen('btm_home').handler.solved_ops
        self.rv.data = self.manager.get_screen('btm_home').handler.get_solved_ops()

    def _update_toolbar(self, *args):
        """Updates the data viewing toolbar based on selections."""
        super(BtmResultsViewer, self)._update_toolbar(self)

        vars_list = ['load', 'charge profile', 'total demand', 'state of charge', 
        # 'total bill', 'total savings'
        ]

        self.vars_button.values = vars_list

    def draw_figure(self, *args):
        self._update_selection()

        if not self._validate_inputs():
            return

        start_time, end_time = self.time_selector.get_inputs()
        plot_type = self.vars_button.text

        results = self.dfs

        if plot_type in ['total bill', 'total savings',]:
            self._reinit_graph(has_legend=False)
        else:
            self._reinit_graph(has_legend=True)

        fig = self.current_fig
        ax = self.current_ax
        ax.clear()

        plt.xticks(rotation=0)
        plt.grid(True)

        if plot_type == 'load':
            for key in results:
                df = results[key]
                ax.plot((df['Pload'])[start_time:end_time], drawstyle='steps-post', label=textwrap.fill(key, 50))

            ax.set_ylabel('kWh')
            ax.set_xlabel('ending hour')
            ax.set_title('Load (kWh)')
        elif plot_type == 'charge profile':
            for key in results:
                df = results[key]

                ax.plot((df['Pcharge'] - df['Pdischarge'])[start_time:end_time], drawstyle='steps-post', label=textwrap.fill(key, 50))

            ax.set_ylabel('kWh')
            ax.set_xlabel('ending hour')
            ax.set_title('Charge Profile (kW)')
        elif plot_type == 'revenue':
            for key in results:
                df = results[key]
                ax.plot(df['revenue'][start_time:end_time], drawstyle='steps-post', label=textwrap.fill(key, 50))

                ax.set_ylabel('$')
                ax.set_xlabel('ending hour')
                ax.set_title('Cumulative Revenue Generated ($)')
                ax.get_yaxis().set_major_formatter(
                    mpl.ticker.FuncFormatter(lambda x, p: format(int(x), ',')))
        elif plot_type == 'state of charge':
            for key in results:
                df = results[key]
                ax.plot(df['state of charge'][start_time:end_time], drawstyle='steps-post', label=textwrap.fill(key, 50))

                ax.set_ylabel('kWh')
                ax.set_xlabel('ending hour')
                ax.set_title('State of Charge (kWh)')
        elif plot_type == 'total demand':
            for key in results:
                df = results[key]
                ax.plot(df['Ptotal'][start_time:end_time], drawstyle='steps-post', label=textwrap.fill(key, 50))

                ax.set_ylabel('kW')
                ax.set_xlabel('ending hour')
                ax.set_title('Total Demand (kW)')
        # elif plot_type == 'total bill':
        #         ixes = range(len(results))
        #         heights = [results[key]['total_bill_with_es'].tail(1) for key in results]

        #         ax.bar(ixes, heights)
        #         labels=[textwrap.fill(' '.join(key.split(' | ')[:3]), 20) for key in results]

        #         ax.set_ylabel('$')
        #         ax.set_title('Total Bill')
        #         ax.grid(False)
        #         plt.xticks(ixes, labels)
        # elif plot_type == 'total savings':
        #         ixes = range(len(results))
        #         heights = [results[key]['total_bill_without_es'].tail(1) - results[key]['total_bill_with_es'].tail(1) for key in results]

        #         ax.bar(ixes, heights)
        #         labels=[textwrap.fill(' '.join(key.split(' | ')[:3]), 20) for key in results]

        #         ax.set_ylabel('$')
        #         ax.set_title('Total Savings')
        #         ax.grid(False)
        #         plt.xticks(ixes, labels)
        # else:
        #     for key in results:
        #         df = results[key]
        #         ax.plot(df[plot_type][start_time:end_time], drawstyle='steps-post', label=textwrap.fill(key, 50))

        #         ax.set_title(plot_type)

        if plot_type in ['total bill', 'total savings']:
            pass
        else:
            ax.legend(bbox_to_anchor=(1.02, 0.5), loc="center left", borderaxespad=0, shadow=False, labelspacing=1.8)

        self.plotbox.children[0].draw()

    def export_png(self):
        """Exports currently displayed figure to .png file in specified location."""
        outdir_root = os.path.join('results', 'btm', 'plots')

        super(BtmResultsViewer, self).export_png(outdir_root)

    def export_csv(self):
        """Exports selected DataFrames to .csv files in specified location."""
        outdir_root = os.path.join('results', 'btm', 'csv')

        super(BtmResultsViewer, self).export_csv(outdir_root)



from __future__ import absolute_import
import math

from pyomo.environ import *


class ExpressionsBlock:
    """Creates blocks for objective and constraint functions and assigns them to the Pyomo model."""

    def set_expressions(self, model):
        """Generates the objective and constraint expressions for model."""
        block_obj = Block()
        model.objectives = block_obj

        block_con = Block()
        model.constraints = block_con

        self._objective_btm(block_obj)
        self._constraints_btm(block_con)
        

    def _objective_btm(self, block):
        eq_objective_btm(block)

    def _constraints_btm(self, block):
        eq_stateofcharge(block)
        eq_stateofcharge_final(block)
        ineq_peak_demand(block)
        ineq_tou_demand(block)
        ineq_nem_xnet(block)

  
def eq_objective_btm(m):
    
    mp = m.parent_block()
    
    _expr = mp.pfpk*mp.flt_dr + sum(mp.ptpk[p]*mp.tou_dr[p] for p in mp.period) +\
    sum(mp.xnet[t]*(mp.tou_er[t]-mp.nem_sr[t])+(mp.pnet[t]+mp.pcha[t]-mp.pdis[t])*mp.nem_sr[t] for t in mp.time)
    
#    _expr = mp.pfpk*mp.flt_dr +\
#    sum(mp.xnet[t]*(mp.tou_er[t]-mp.nem_sr[t])+(mp.pnet[t]+mp.pcha[t]-mp.pdis[t])*mp.nem_sr[t] for t in mp.time)
#    
    mp.objective_expr += _expr
    m.objective_rt = Expression(expr=_expr)


def eq_stateofcharge(m):
    """Definition of state of charge"""
    mp = m.parent_block()
   
    def _eq_stateofcharge(_m, t):
        if t==0:
            spre=mp.State_of_charge_init*mp.Energy_capacity
        else:
            spre=mp.s[t-1]
        return mp.Self_discharge_efficiency * spre + mp.Round_trip_efficiency * mp.pcha[t] \
            - mp.pdis[t] == mp.s[t]       

    m.stateofcharge = Constraint(mp.time, rule=_eq_stateofcharge)

def eq_stateofcharge_final(m):
    """Requires the final state of charge of the energy storage device to equal its initial value."""
    mp = m.parent_block()
    T  = mp.nhr-1
    def _eq_stateofcharge_final(_m, t):
        return mp.s[T] == mp.State_of_charge_init*mp.Energy_capacity
    m.stateofcharge_final = Constraint(mp.time, rule=_eq_stateofcharge_final)


def ineq_peak_demand(m):
    """Requires all net power at time t less the peak demand"""
    mp = m.parent_block()
    def _ineq_peak_demand(_m, t):
        return mp.pnet[t]+mp.pcha[t]-mp.pdis[t]-mp.pfpk<=0
    m.peak_demand = Constraint(mp.time, rule=_ineq_peak_demand)

def ineq_tou_demand(m):
    """Requires all net power at time t period p less the peak demand of period p"""
    mp = m.parent_block()
    def _ineq_tou_demand(_m, p, t):
        return mp.mask_ds[p][t]*(mp.pnet[t]+mp.pcha[t]-mp.pdis[t])<=mp.ptpk[p]
    m.tou_demand = Constraint(mp.period, mp.time, rule=_ineq_tou_demand)
    
def ineq_nem_xnet(m):
    """Requires all net power at time t less the peak demand"""
    mp = m.parent_block()
    def _ineq_nem_xnet(_m, t):
        return mp.pnet[t]+mp.pcha[t]-mp.pdis[t]<= mp.xnet[t]
    m.nem_xnet = Constraint(mp.time, rule=_ineq_nem_xnet)


from __future__ import absolute_import

import json
import collections
import logging
import copy
import calendar
from functools import partial

import numpy as np

from kivy.clock import Clock
from kivy.uix.screenmanager import Screen, ScreenManager
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.gridlayout import GridLayout
from kivy.uix.textinput import TextInput
from kivy.properties import ObjectProperty, BooleanProperty, StringProperty, DictProperty
from kivy.app import App
from kivy.animation import Animation
from kivy.uix.stencilview import StencilView

from es_gui.resources.widgets.common import InputError, WarningPopup, MyPopup, ValuationParameterRow, ValuationParameterWidget, ValuationRunCompletePopup, RecycleViewRow, FADEIN_DUR, fade_in_animation
from es_gui.tools.valuation.valuation_optimizer import BadParameterException
from es_gui.apps.data_manager.data_manager import DataManagerException


class ValuationBatchRunScreenManager(ScreenManager):
    def __init__(self, **kwargs):
        super(ValuationBatchRunScreenManager, self).__init__(**kwargs)

        # Add new screens here.
        self.add_widget(BatchRunDataScreen(name='data'))
        self.add_widget(BatchRunParamScreen(name='params'))


class BatchRunScreen(Screen):
    def on_enter(self):
        # Change the navigation bar title.
        ab = self.manager.nav_bar
        ab.build_valuation_batch_nav_bar()
        ab.set_title('Batch Runs')

    def _generate_requests(self):
        data_screen = self.batch_sm.get_screen('data')
        params_screen = self.batch_sm.get_screen('params')

        rv_selected, months, iso, rev_streams, node = data_screen.get_inputs()
        param_settings = params_screen.get_inputs()

        requests = {'iso': iso,
                    'market type': rev_streams,
                    'months': months,
                    'node id': node['nodeid'],
                    'param set': param_settings
                    }

        return requests

    def run_batch(self):
        try:
            requests = self._generate_requests()
        except ValueError as e:
            popup = WarningPopup()
            popup.popup_text.text = str(e)
            popup.open()
        except InputError as e:
            popup = WarningPopup()
            popup.popup_text.text = str(e)
            popup.open()
        else:
            solver_name = App.get_running_app().config.get('optimization', 'solver')

            handler = self.manager.get_screen('valuation_home').handler
            handler.solver_name = solver_name

            try:
                solved_ops, handler_status = handler.process_requests(requests)
            except BadParameterException as e:
                popup = WarningPopup()
                popup.popup_text.text = str(e)
                popup.open()
            else:
                self.completion_popup = BatchRunCompletePopup()
                self.completion_popup.view_results_button.bind(on_release=self._go_to_view_results)

                if len(handler_status) > 0:
                    if solved_ops:
                        # At least one model solved successfully.
                        self.completion_popup.title = "Success!*"
                        self.completion_popup.popup_text.text = '\n'.join([
                            'All finished, but we found these issues:',
                        ]
                        + list(handler_status)
                        )
                    else:
                        # No models solved successfully.
                        self.completion_popup.title = "Hmm..."
                        self.completion_popup.popup_text.text = '\n'.join([
                            'Unfortunately, none of the models were able to be solved. We found these issues:',
                        ]
                        + list(handler_status)
                        )

                self.completion_popup.open()            

    def _go_to_view_results(self, *args):
        self.manager.nav_bar.go_to_screen('valuation_results_viewer')
        self.completion_popup.dismiss()


class BatchRunParamScreen(Screen):
    iso = StringProperty('')
    param_to_attr = dict()

    def on_pre_enter(self):
        self.iso = self.manager.get_screen('data').iso

    def on_iso(self, instance, value):
        while len(self.param_widget.children) > 0:
            for widget in self.param_widget.children:
                if isinstance(widget, ValuationParameterRow):
                    self.param_widget.remove_widget(widget)
    
        if self.iso:
            try:
                self.param_widget.build(self.iso)

                data_manager = App.get_running_app().data_manager
                MODEL_PARAMS = data_manager.get_valuation_model_params(value)

                self.param_to_attr = {param['name']: param['attr name']
                                      for param in MODEL_PARAMS}
                self.param_sweep_spinner.values = ['none',] + list(self.param_to_attr.keys())
            except KeyError:
                pass

    def _disable_text_input(self, text):
        # Disables the text input field of the parameter selected for a parameter sweep.
        for row_widget in self.param_widget.children:
            row_widget.text_input.disabled = False

        attr_name = self.param_to_attr.get(text, '')

        if attr_name:
            param_widget_row = getattr(self.param_widget, attr_name, None)
            param_widget_row.text_input.disabled = True

    def _validate_inputs(self):
        if len(self.param_widget.children) == 0:
            raise InputError('Please specify the simulation parameters.')

        param_dict = self.param_widget.get_inputs(use_hint_text=True)

        # If a parameter sweep is specified, check all input fields.
        param_sweep_name = self.param_sweep_spinner.text

        if self.param_to_attr.get(param_sweep_name, False):
            param_sweep_attr_name = self.param_to_attr[param_sweep_name]

            try:
                param_min = float(self.param_min_input.text)
                param_max = float(self.param_max_input.text)
                param_step = int(self.param_step_input.text)
            except ValueError:
                raise InputError('All range input fields must be populated when specifying a parameter sweep.')
            else:
                if param_max < param_min:
                    raise InputError('Parameter sweep minimum must be less than or equal to the maximum.')
            
            # Values cannot be negative.
            if any([param_min < 0, param_max < 0]):
                raise InputError('"{0}" cannot be negative.'.format(param_sweep_name))

            # Percentages cannot exceed 100%.
            if param_sweep_attr_name in {'Self_discharge_efficiency', 'Round_trip_efficiency', 'State_of_charge_init', 'State_of_charge_min', 'State_of_charge_max', 'Reserve_reg_min', 'Reserve_reg_max',} and any([param_min > 100, param_max > 100]):
                raise InputError('"{0}" cannot exceed 100%.'.format(param_sweep_name))
            
            # State of charge initial, min, and max sweeps must comply with inequalities.
            if param_sweep_attr_name == 'State_of_charge_min':
                if not all([param_min < param_dict['State_of_charge_init'], param_max <= param_dict['State_of_charge_init']]):
                    raise InputError('The parameter sweep range for "{0}" must be entirely less than the initial state of charge.'.format(param_sweep_name))
            elif param_sweep_attr_name == 'State_of_charge_max':
                if not all([param_min >= param_dict['State_of_charge_init'], param_max > param_dict['State_of_charge_init']]):
                    raise InputError('The parameter sweep range for "{0}" must be entirely greater than the initial state of charge.'.format(param_sweep_name))
            elif param_sweep_attr_name == 'State_of_charge_init':
                if not all([
                    param_dict['State_of_charge_max'] > param_min,
                    param_dict['State_of_charge_max'] >= param_max,
                    param_dict['State_of_charge_min'] <= param_min,
                    param_dict['State_of_charge_min'] < param_max,
                    ]):
                    raise InputError('The parameter sweep range for "{0}" must be entirely between the minimum and maximum state of charge values.'.format(param_sweep_name))
        else:
            if any([self.param_min_input.text, self.param_max_input.text, self.param_step_input.text]):
                raise InputError('Did you forget to provide a parameter for the parameter sweep? Numbers for the sweep range were provided but no parameter was given.')

    def get_inputs(self):
        self._validate_inputs()

        base_param_dict = self.param_widget.get_inputs()

        param_settings = []

        param_sweep_name = self.param_sweep_spinner.text

        if self.param_to_attr.get(param_sweep_name, False):
            param_min = float(self.param_min_input.text)
            param_max = float(self.param_max_input.text)
            param_step = int(self.param_step_input.text)
            param_sweep_range = np.linspace(param_min, param_max, num=param_step)

            for param_sweep_value in param_sweep_range:
                base_copy = copy.deepcopy(base_param_dict)
                base_copy[self.param_to_attr[param_sweep_name]] = float(param_sweep_value)
                param_settings.append(base_copy)
        else:
            param_settings.append(base_param_dict)

        return param_settings


class BatchRunDataScreen(Screen):
    iso = StringProperty('')
    rev_streams = StringProperty('')
    node = DictProperty()

    def __init__(self, **kwargs):
        super(BatchRunDataScreen, self).__init__(**kwargs)

        BatchRVNodeEntry.host_screen = self
    
    def _get_iso_options(self):
        try:
            data_manager = App.get_running_app().data_manager
        except AttributeError:
            pass
        else:
            iso_options = data_manager.get_markets()
            self.iso_select.values = iso_options

    def _reset_screen(self):
        self.revstreams_select.opacity = 0.05
        self.months_select_bx.opacity = 0.05
        self.node_select_bx.opacity = 0.05

        # Deselects all RV selections.
        self.month_rv.deselect_all_nodes()
        self.node_rv.deselect_all_nodes()

        # Resets properties.
        self.rev_streams = ''
        self.node = {}

    def _iso_selected(self):
        self.iso = self.iso_select.text
        self.revstreams_select.text = 'Select revenue streams'

        try:
            # Get available pricing nodes.
            data_manager = App.get_running_app().data_manager
            nodes = data_manager.get_nodes(self.iso)

            node_options = [{'name': node[1],
                             'nodeid': node[0]} for node in nodes.items()]

            self.node_rv.data = node_options
            self.node_rv.unfiltered_data = node_options
        except KeyError:
            # The ISO selected has no valid nodes.
            node_options = [{}]
        else:
            # Enable node selector.
            Clock.schedule_once(partial(fade_in_animation, self.node_select_bx), 0)

    def on_iso(self, instance, value):
        logging.info('ValuationBatch: ISO changed to {0}.'.format(value))

        self._reset_screen()

    def _revstreams_selected(self):
        try:
            data_manager = App.get_running_app().data_manager
            market_models_available = data_manager.get_valuation_revstreams(self.iso, self.node['nodeid'])
            self.rev_streams = market_models_available[self.revstreams_select.text]['market type']
        except KeyError:
            self.rev_streams = ''
        else:
            # Get available historical data.
            historical_data_all = data_manager.get_historical_data_options(self.iso, self.node['nodeid'], self.revstreams_select.text)
            historical_data_opts = []

            for (year, month_list) in historical_data_all.items():
                month_options = [{'name': '{month} {year}'.format(month=calendar.month_name[int(month)], year=year),
                'month': month, 'year': year} for month in month_list]
                historical_data_opts.extend(month_options)
            
            self.month_rv.data = historical_data_opts
            Clock.schedule_once(partial(fade_in_animation, self.months_select_bx), 0)

    def on_rev_streams(self, instance, value):
        if value:
            logging.info('ValuationBatch: Revenue streams changed to {0}.'.format(value))
        else:
            logging.info('ValuationBatch: Revenue streams selection reset.')

    def on_node(self, instance, value):
        try:
            logging.info('ValuationBatch: Pricing node changed to {0}.'.format(self.node['name']))
        except KeyError:
            logging.info('ValuationBatch: Pricing node selection reset.')
        else:
            # Get available revenue streams.
            try:
                data_manager = App.get_running_app().data_manager
                market_models_available = data_manager.get_valuation_revstreams(self.iso, value['nodeid'])
                self.revstreams_select.values = market_models_available.keys()
            except (KeyError, DataManagerException):
                self.revstreams_select.values = []
            else:
                # Enable revenue stream selector and reset it.
                Clock.schedule_once(partial(fade_in_animation, self.revstreams_select), 0)
                self.revstreams_select.text = 'Select revenue streams'
                self.revstreams_select.disabled = False
            
            # Reset historical data options.
            self._revstreams_selected()
            self.month_rv.deselect_all_nodes()
            self.month_rv.data = []

    def _validate_inputs(self):
        if self.iso_select.text == 'Select market area':
            raise (InputError('No market area selected.'))

        if self.revstreams_select.text == 'Select revenue streams':
            raise (InputError('No revenue streams selected.'))

        if not self.node:
            raise (InputError('No pricing node selected.'))

        rv_selected = [self.month_rv.data[selected_ix] for selected_ix in self.month_rv.layout_manager.selected_nodes]

        if not rv_selected:
            raise (InputError('No data selected.'))

    def get_inputs(self):
        self._validate_inputs()

        # Retrieve selected months from recycle view.
        rv_selected = [self.month_rv.data[selected_ix] for selected_ix in self.month_rv.layout_manager.selected_nodes]

        months = [(selected_entry['month'], selected_entry['year']) for
                  selected_entry in rv_selected]

        return rv_selected, months, self.iso, self.rev_streams, self.node


class BatchRVNodeEntry(RecycleViewRow):
    host_screen = None

    def apply_selection(self, rv, index, is_selected):
        """
        Respond to the selection of items in the view.
        """
        super(BatchRVNodeEntry, self).apply_selection(rv, index, is_selected)

        if is_selected:
            self.host_screen.node = rv.data[self.index]
        # else:
        #     self.host_screen.node = ''


class BatchRunCompletePopup(ValuationRunCompletePopup):
    def __init__(self, **kwargs):
        super(BatchRunCompletePopup, self).__init__(**kwargs)

        self.popup_text.text = 'Your specified batch runs have been completed.'



"""Example for running valuation simulations."""
from __future__ import absolute_import

# import sys

# sys.path.append("C:/work/quest-library/")
import logging
from datetime import datetime
import calendar
import pyutilib
import matplotlib.pyplot as plt

import quest_eras.quest_library.utilities.pomo.valuation.caiso.valuation_optimizer as caiso_op
from quest_eras.quest_library.utilities.pomo.valuation.caiso.valuation_dms import (
    ValuationDMSCAISO,
)
import quest_eras.quest_library.utilities.pomo.valuation.ercot.valuation_optimizer as ercot_op
from quest_eras.quest_library.utilities.pomo.valuation.ercot.valuation_dms import (
    ValuationDMSERCOT,
)
import quest_eras.quest_library.utilities.pomo.valuation.isone.valuation_optimizer as isone_op
from quest_eras.quest_library.utilities.pomo.valuation.isone.valuation_dms import (
    ValuationDMSISONE,
)
import quest_eras.quest_library.utilities.pomo.valuation.miso.valuation_optimizer as miso_op
from quest_eras.quest_library.utilities.pomo.valuation.miso.valuation_dms import (
    ValuationDMSMISO,
)
import quest_eras.quest_library.utilities.pomo.valuation.nyiso.valuation_optimizer as nyiso_op
from quest_eras.quest_library.utilities.pomo.valuation.nyiso.valuation_dms import (
    ValuationDMSNYISO,
)
import quest_eras.quest_library.utilities.pomo.valuation.pjm.valuation_optimizer as pjm_op
from quest_eras.quest_library.utilities.pomo.valuation.pjm.valuation_dms import (
    ValuationDMSPJM,
)
import quest_eras.quest_library.utilities.pomo.valuation.spp.valuation_optimizer as spp_op
from quest_eras.quest_library.utilities.pomo.valuation.spp.valuation_dms import (
    ValuationDMSSPP,
)


class ValuationOptimizerHandler:
    """A handler for creating and solving ValuationOptimizer instances as requested."""

    solved_ops = []

    def __init__(self, solver_name, data_path, **kwargs):
        self._solver_name = solver_name
        self.caiso_dms = ValuationDMSCAISO(data_path, **kwargs)
        self.ercot_dms = ValuationDMSERCOT(data_path, **kwargs)
        self.isone_dms = ValuationDMSISONE(data_path, **kwargs)
        self.miso_dms = ValuationDMSMISO(data_path, **kwargs)
        self.nyiso_dms = ValuationDMSNYISO(data_path, **kwargs)
        self.pjm_dms = ValuationDMSPJM(data_path, **kwargs)
        self.spp_dms = ValuationDMSSPP(data_path, **kwargs)

    @property
    def solver_name(self):
        """The name of the solver for Pyomo to call."""
        return self._solver_name

    @solver_name.setter
    def solver_name(self, value):
        self._solver_name = value

    def process_requests(self, requests, *args, **kwargs):
        """Generates and solves ValuationOptimizer models based on the given requests."""
        iso = requests["iso"]
        market_type = requests["market type"]
        node_id = str(requests["node id"])
        node_name = self.caiso_dms.get_node_name(node_id, iso)
        param_set = requests.get("param set", [None])

        solved_requests = []

        handler_status = set()

        for month, year in requests["months"]:
            param_set_iterator = iter(param_set)
            continue_param_loop = True

            while continue_param_loop:
                try:
                    params = next(param_set_iterator)
                except StopIteration:
                    break

                if iso == "PJM":
                    from quest_eras.quest_library.utilities.pomo.valuation.pjm.valuation_optimizer import (
                        IncompatibleDataException,
                    )  # find a better solution

                    lmp_da, MR, RA, RD, RegCCP, RegPCP = self.pjm_dms.get_pjm_data(
                        year, month, node_id
                    )
                    op = pjm_op.ValuationOptimizerPJM(market_type=market_type)

                    op.price_electricity = lmp_da
                    op.mileage_mult = MR
                    op.price_regulation = RegCCP
                    op.price_reg_service = RegPCP
                elif iso == "ERCOT":
                    from quest_eras.quest_library.utilities.pomo.valuation.ercot.valuation_optimizer import (
                        IncompatibleDataException,
                    )  # find a better solution

                    lmp_da, rd, ru = self.ercot_dms.get_ercot_data(
                        year, month, node_name
                    )
                    op = ercot_op.ValuationOptimizerERCOT(market_type=market_type)

                    op.price_electricity = lmp_da
                    op.price_reg_up = ru
                    op.price_reg_down = rd
                elif iso == "MISO":
                    from quest_eras.quest_library.utilities.pomo.valuation.miso.valuation_optimizer import (
                        IncompatibleDataException,
                    )  # find a better solution

                    lmp_da, regMCP = self.miso_dms.get_miso_data(year, month, node_name)
                    op = miso_op.ValuationOptimizerMISO(market_type=market_type)

                    op.price_electricity = lmp_da
                    op.price_regulation = regMCP
                elif iso == "ISONE":
                    from quest_eras.quest_library.utilities.pomo.valuation.isone.valuation_optimizer import (
                        IncompatibleDataException,
                    )  # find a better solution

                    daLMP, RegCCP, RegPCP, miMULT = self.isone_dms.get_isone_data(
                        year, month, node_id
                    )
                    op = isone_op.ValuationOptimizerISONE(market_type=market_type)

                    op.price_electricity = daLMP
                    op.price_regulation = RegCCP
                    op.price_reg_service = RegPCP
                    op.mileage_mult = miMULT
                elif iso == "NYISO":
                    from quest_eras.quest_library.utilities.pomo.valuation.nyiso.valuation_optimizer import (
                        IncompatibleDataException,
                    )  # find a better solution

                    lbmp_da, rcap_da = self.nyiso_dms.get_nyiso_data(
                        year, month, node_id
                    )
                    op = nyiso_op.ValuationOptimizerNYISO(market_type=market_type)

                    op.price_electricity = lbmp_da
                    op.price_regulation = rcap_da
                elif iso == "SPP":
                    from quest_eras.quest_library.utilities.pomo.valuation.spp.valuation_optimizer import (
                        IncompatibleDataException,
                    )  # find a better solution

                    lmp_da, mcpru_da, mcprd_da = self.spp_dms.get_spp_data(
                        year, month, node_name
                    )
                    op = spp_op.ValuationOptimizerSPP(market_type=market_type)

                    op.price_electricity = lmp_da
                    op.price_reg_up = mcpru_da
                    op.price_reg_down = mcprd_da
                elif iso == "CAISO":
                    from quest_eras.quest_library.utilities.pomo.valuation.caiso.valuation_optimizer import (
                        IncompatibleDataException,
                    )  # find a better solution

                    (
                        lmp_da,
                        aspru_da,
                        asprd_da,
                        asprmu_da,
                        asprmd_da,
                        rmu_mm,
                        rmd_mm,
                        rmu_pacc,
                        rmd_pacc,
                    ) = self.caiso_dms.get_caiso_data(year, month, node_name)
                    op = caiso_op.ValuationOptimizerCAISO(market_type=market_type)

                    op.price_electricity = lmp_da
                    op.price_reg_up = aspru_da
                    op.price_reg_down = asprd_da
                    op.price_reg_serv_up = asprmu_da
                    op.price_reg_serv_down = asprmd_da
                    op.mileage_mult_ru = rmu_mm
                    op.mileage_mult_rd = rmd_mm
                    op.perf_score_ru = (
                        rmu_pacc  # TODO: give the option to the user to override this
                    )
                    op.perf_score_rd = rmd_pacc
                else:
                    logging.error("ValOp Handler: Invalid ISO provided.")
                    raise ValueError(
                        "Invalid ISO provided to ValuationOptimizer handler."
                    )

                if params:
                    op.set_model_parameters(**params)
                else:
                    continue_param_loop = False

                try:
                    solved_op = self._solve_model(op)
                except pyutilib.common._exceptions.ApplicationError as e:
                    logging.error("Op Handler: {error}".format(error=e))

                    if "No executable found" in e.args[0]:
                        # Could not locate solver executable
                        handler_status.add(
                            "* The executable for the selected solver could not be found; please check your installation."
                        )
                    else:
                        handler_status.add(
                            "* ({0} {1}) {2}. The problem may be infeasible.".format(
                                month, year, e.args[0]
                            )
                        )
                except IncompatibleDataException as e:
                    # Data exception raised by ValuationOptimizer
                    logging.error(e)
                    handler_status.add(
                        "* ({0} {1}) The time series data has mismatched sizes.".format(
                            month, year
                        )
                    )
                except AssertionError as e:
                    # An optimal solution could not be found as reported by the solver
                    logging.error("Op Handler: {error}".format(error=e))
                    handler_status.add(
                        "* ({0} {1}) An optimal solution could not be found; the problem may be infeasible.".format(
                            month, year
                        )
                    )
                else:
                    solved_op = self._save_to_solved_ops(
                        solved_op, iso, market_type, node_name, year, month, params
                    )
                    solved_requests.append(solved_op)

        logging.info("Op Handler: Finished processing requested jobs.")
        return solved_requests, handler_status

    def _solve_model(self, op):
        op.solver = self.solver_name
        op.run()

        return op

    @staticmethod
    def _save_to_solved_ops(op, iso, market_type, node_name, year, month, param_set):
        # time_finished = datetime.now().strftime('%A, %B %d, %Y %H:%M:%S')
        time_finished = datetime.now().strftime("%b %d, %Y %H:%M:%S")
        name = " | ".join(
            [
                time_finished,
                str(node_name),
                str(year),
                calendar.month_abbr[int(month)],
                repr(param_set),
            ]
        )

        results_dict = {}

        results_dict["name"] = name
        results_dict["optimizer"] = op
        results_dict["iso"] = iso
        results_dict["market type"] = market_type
        results_dict["year"] = year
        results_dict["month"] = calendar.month_name[int(month)]
        results_dict["node"] = node_name
        if param_set:
            results_dict["params"] = param_set
        results_dict["time"] = time_finished
        results_dict["label"] = " ".join(
            [
                str(node_name),
                str(year),
                calendar.month_abbr[int(month)],
                repr(param_set),
            ]
        )

        ValuationOptimizerHandler.solved_ops.append(results_dict)

        return (name, op)

    def get_solved_ops(self):
        """Returns the list of solved Optimizer objects in reverse chronological order."""
        return_list = reversed(self.solved_ops)

        return return_list


# if __name__ == "__main__":
#     with open("valuation_optimizer.log", "w"):
#         pass

#     logging.basicConfig(
#         filename="valuation_optimizer.log",
#         format="[%(levelname)s] %(asctime)s: %(message)s",
#         level=logging.INFO,
#     )

#     params = [
#         {
#             "Power_rating": 2,  # MW
#             "Energy_capacity": 4,  # MWh
#             "Self_discharge_efficiency": 1,
#             "R": 0,  # Interest rate [hour^-1]
#             "Round_trip_efficiency": 0.85,
#             "Reserve_reg_min": 0,  # Fraction of q_reg bid to increase state of charge minimum by.
#             "Reserve_reg_max": 1,  # Fraction of q_reg bid to decrease state of charge maximum by.
#             "State_of_charge_min": 0,  # Fraction of energy capacity representing the minimum state of charge.
#             "State_of_charge_max": 1,  # Fraction of energy capacity representing the maximum state of charge.
#             "State_of_charge_init": 0.5,  # Initial state of charge as a fraction of energy capacity.
#         }
#     ]

#     requests_PJM = {
#         "iso": "PJM",
#         "node id": 1,  # PJM-rTO
#         "market type": ["arbitrage", "pfp"],
#     }

#     requests_MISO = {
#         "iso": "MISO",
#         "node id": "MICHIGAN.HUB",
#         "market type": ["arbitrage", "pfp"],
#     }

#     requests_ISONE = {
#         "iso": "ISONE",
#         "node id": 4004,  # CONNECTICUT HUB
#         "market type": ["arbitrage", "pfp"],
#     }

#     requests_NYISO = {
#         "iso": "NYISO",
#         "node id": 61761,  # NYC ZONE
#         "market type": ["arbitrage", "pfp"],
#     }

#     requests_CAISO = {
#         "iso": "CAISO",
#         "node id": "TH_SP15_GEN-APND",  # SOUTH HUB
#         "market type": ["arbitrage", "pfp"],
#     }

#     requests_SPP = {
#         "iso": "SPP",
#         "node id": "AECI",
#         "market type": ["arbitrage", "pfp"],
#     }

#     requests_ERCOT = {
#         "iso": "ERCOT",
#         "node id": "HB_HOUSTON",
#         "market type": ["arbitrage", "pfp"],
#     }

#     months = [(1, 2018), (2, 2018), (3, 2018)]
#     requests = [
#         requests_PJM,
#         requests_MISO,
#         requests_ISONE,
#         requests_NYISO,
#         requests_CAISO,
#         requests_SPP,
#         requests_ERCOT,
#     ]
#     handler = ValuationOptimizerHandler(
#         solver_name="glpk",
#         data_path="apps/valuation/data/",
#         save_name="valuation_dms.p",
#     )

#     for request in requests:
#         request["param set"] = params
#         request["months"] = months
#         market_types = request["market type"]
#         for market_type in market_types:
#             request["market type"] = market_type

#             handler.process_requests(request)

#     solved_ops = handler.get_solved_ops()

#     results_dict = {
#         "PJM": [],
#         "MISO": [],
#         "ISONE": [],
#         "NYISO": [],
#         "CAISO": [],
#         "SPP": [],
#         "ERCOT": [],
#     }

#     for result in solved_ops:
#         if result["iso"] == "PJM":
#             results_dict["PJM"].append(result)
#         elif result["iso"] == "MISO":
#             results_dict["MISO"].append(result)
#         elif result["iso"] == "ISONE":
#             results_dict["ISONE"].append(result)
#         elif result["iso"] == "NYISO":
#             results_dict["NYISO"].append(result)
#         elif result["iso"] == "CAISO":
#             results_dict["CAISO"].append(result)
#         elif result["iso"] == "SPP":
#             results_dict["SPP"].append(result)
#         elif result["iso"] == "ERCOT":
#             results_dict["ERCOT"].append(result)

#     fig_ls = []
#     for iso in results_dict:
#         fig, ax = plt.subplots()
#         arb = []
#         pfp = []
#         for result in results_dict[iso]:
#             if result["market type"] == "arbitrage":
#                 arb.append((result["month"], result["optimizer"].gross_revenue))
#             else:
#                 pfp.append((result["month"], result["optimizer"].gross_revenue))

#         ax.bar(
#             x=[date[0] for date in arb],
#             height=[date[1] for date in arb],
#             width=-0.5,
#             align="edge",
#             color=["blue"],
#         )
#         ax.bar(
#             x=[date[0] for date in pfp],
#             height=[date[1] for date in pfp],
#             width=0.5,
#             align="edge",
#             color=["orange"],
#         )
#         ax.legend()
#         ax.set_title("{} Gross Revenue 2018".format(iso))
#         ax.set_ylabel("$")
#         fig_ls.append((fig, ax))
#         fig_ls.append((fig, ax))
#         fig_ls.append((fig, ax))



import os
import re
import numpy as np
import pandas as pd
from es_gui.apps.tech_selection import fFeasibility, fPlots, fAux

from kivy.app import App

# Define some auxiliary functions
normalize_by_max = lambda x: x/x.max()
normalize_by_target = lambda x, tc: tc/(x+tc)


def perform_tech_selection(selections, target_cost_kWh, target_cost_kW):
    """Perform energy storage technology selection based on the user inputs."""

    data_manager = App.get_running_app().data_manager
    app_data = data_manager.get_applications_db()

    # Unpack user selections (input parameters)
    grid_location = selections['location']
    application = selections['application']
    system_size = selections['system_size']
    discharge_duration = float('.'.join(re.findall(r'\d+', selections['discharge_duration'])))
    app_type = selections['app_type']
    target_cost = target_cost_kW if app_type == 'Power' else target_cost_kWh

    # Read databases
    tech_data = data_manager.get_techs_db()

    tech_data.rename(columns={'Feas. score for residential': 'Score for BTM: residential',
                              'Feas. score for industrial': 'Score for BTM: commercial/industrial',
                              'Feas. score for distribution': 'Score for Distribution',
                              'Feas. score for transmission': 'Score for Transmission/central'}, inplace=True)

    # 1st filter: compatibility between grid location and ES techs
    feas_location = list(set(fFeasibility.isFeasibleTech_Location(tech_data, grid_location)))

    # 2nd filter: minimum application requirements
    feas_duration = list(set(fFeasibility.isFeasibleTech_Duration(tech_data, discharge_duration)))
    feas_response_time = list(set(fFeasibility.isFeasibleTech_ResponseTime(
        tech_data, app_data.loc[application, 'Minimum required response time'])))
    feas_electric_output = list(set(fFeasibility.isFeasibleTech_ElectricOutput(
        tech_data, app_data.loc[application, 'Requires electric output'])))

    # Combine feasibility results from all filters (each entry is the boolean True or False)
    all_feasibility = pd.DataFrame(index=np.unique(tech_data['Storage technology (short name)'].values))
    all_feasibility['Grid location'] = [value in feas_location for value in all_feasibility.index]
    all_feasibility['Application requirements'] = [
        value in feas_duration and value in feas_response_time and value in feas_electric_output
        for value in all_feasibility.index]
    all_feasibility['Feasible?'] = all_feasibility.all(axis='columns')

    # For each feasible technology, consider only the entry with the lowest discharge duration that satisfies the requirements
    aa = pd.DataFrame(tech_data.loc[tech_data['Discharge duration (hours)']>=discharge_duration]).groupby(by='Short abbreviation')
    feas_techs_lowest_duration = []
    for _, grp in aa:
        feas_techs_lowest_duration.append(grp.index.tolist()[0])

    # Subset the technology database to contain only the feasible technologies
    tech_data_lowest_duration = pd.DataFrame(tech_data.loc[feas_techs_lowest_duration])
    tech_data_lowest_duration.reset_index(inplace=True, drop=True)
    tech_data_lowest_duration.set_index('Storage technology (short name)', inplace=True)

    # Compute 'Application score' for all feasible technologies (based on discharge duration, cycle life, and efficiency)
    all_app_scores = pd.DataFrame(index=all_feasibility.index)
    all_app_scores['Score for duration'] = tech_data_lowest_duration[['Discharge duration (hours)']].apply(normalize_by_max)
    all_app_scores['Score for cycle life'] = tech_data_lowest_duration[['Cycle life (# of cycles)']].apply(normalize_by_max)
    all_app_scores['Score for efficiency'] = tech_data_lowest_duration[['Round-trip efficiency (%)']].apply(normalize_by_max)
    all_app_scores.fillna(value=0, inplace=True)
    all_app_scores['Application score'] = fAux.geom_mean(all_app_scores)

    # Compute 'Total score' for all feasible technologies (based on application, location, cost, and maturity)
    all_final_scores = pd.DataFrame(index=all_feasibility.index)
    all_final_scores['Application score'] = all_app_scores['Application score']
    all_final_scores['Location score'] = tech_data_lowest_duration[f'Score for {grid_location}']
    all_final_scores['Cost score'] = compute_cost_scores(tech_data_lowest_duration, system_size, app_type, target_cost)
    all_final_scores['Maturity score'] = tech_data_lowest_duration['Tech readiness score']
    all_final_scores.fillna(value=0, inplace=True)
    all_final_scores['Total score'] = all_final_scores['Maturity score'] *\
        fAux.geom_mean(all_final_scores[['Application score', 'Location score', 'Cost score']])
    all_final_scores.sort_values(by=['Total score', 'Application score', 'Location score', 'Cost score', 'Maturity score'],
                                 inplace=True)

    # Plot: feasibility heatmap
    fig = fPlots.plot_table_feasibility(all_feasibility, figsize=(1.4, 1),
                                        xticklabels=['Grid location', 'Application\nrequirements', 'Feasible?'])
    fig.savefig(os.path.join('results', 'tech_selection', 'plot_feasibility.png'))

    # Plot: final feasibility scores
    fig = fPlots.plot_ranking_techs(all_final_scores)
    fig.savefig(os.path.join('results', 'tech_selection', 'plot_ranking.png'))

    return all_feasibility, all_final_scores


def compute_cost_scores(techDB, system_size, app_type, target_cost):
    """Compute cost score for each technology."""
    if app_type == 'Power':
        cost_scores = techDB[f'Cost at {system_size} ($/kW)'].apply(normalize_by_target, tc=target_cost).fillna(value=0)
    elif app_type == 'Energy':
        kWh_costs = techDB[f'Cost at {system_size} ($/kW)']/techDB['Discharge duration (hours)']
        cost_scores = kWh_costs.apply(normalize_by_target, tc=target_cost).fillna(value=0)

    return cost_scores


"""Optimizer abstract class."""

from abc import ABCMeta, abstractmethod
import logging
import pyutilib
import pyomo.environ as pom
from pyomo.opt import TerminationCondition
from six import with_metaclass

# from pyomo.environ import *


class Optimizer(with_metaclass(ABCMeta)):
    """Abstract base class for Pyomo ConcreteModel optimization framework."""

    def __init__(self, solver="glpk"):
        self._model = pom.ConcreteModel()
        self._solver = solver

        self._results = None

    @property
    def model(self):
        """Create a Pyomo ConcreteModel."""
        return self._model

    @property
    def solver(self):
        """Specify the name of the solver for Pyomo to use."""
        return self._solver

    @property
    def results(self):
        """Return results in a DataFrame containing indices, decision variables, model parameters and derived quantities."""
        return self._results

    @abstractmethod
    def _set_model_param(self):
        """Assign model parameters and their default values to the model."""
        pass

    @abstractmethod
    def _set_model_var(self):
        """Initialize model decision variables for the model."""
        pass

    @abstractmethod
    def instantiate_model(self):
        """Instantiate the model and assign Optimizer attributes to model attributes."""
        pass

    @abstractmethod
    def populate_model(self):
        """Set model parameters, variables, and an ExpressionsBlock object for defining objectives and constraints."""
        pass

    def solve_model(self):
        """Solve the model using the specified solver."""
        if self.solver == "neos":
            opt = pom.SolverFactory("cbc")
            solver_manager = pom.SolverManagerFactory("neos")
            results = solver_manager.solve(self.model, opt=opt)
        else:
            solver = pom.SolverFactory(self.solver)
            results = solver.solve(self.model, tee=False, keepfiles=False)

        assert results.solver.termination_condition.key == "optimal"

        self._process_results()

    @abstractmethod
    def _process_results(self):
        """Compute derived quantities of interest and creating the results DataFrame."""
        pass

    @abstractmethod
    def get_results(self):
        """Return the results DataFrame plus any other quantities of interest."""
        pass

    def run(self):
        """Instantiate, create, and solve the optimizer model based on provided information.

        Use if no steps are needed between constructing the model and solving it.
        """
        self.instantiate_model()
        self.populate_model()

        if self.solver == "neos":
            opt = pom.SolverFactory("cbc")
            solver_manager = pom.SolverManagerFactory("neos")
            results = solver_manager.solve(self.model, opt=opt)
        else:
            solver = pom.SolverFactory(self.solver)

            try:
                solver.available()
            except pyutilib.common._exceptions.ApplicationError as e:
                logging.error("Optimizer: {error}".format(error=e))
            else:
                results = solver.solve(self.model, tee=True, keepfiles=False)

        try:
            assert results.solver.termination_condition == TerminationCondition.optimal
        except AssertionError:
            logging.error(
                "Optimizer: An optimal solution could not be obtained. (solver termination condition: {0})".format(
                    results.solver.termination_condition.key
                )
            )
            raise (
                AssertionError(
                    "An optimal solution could not be obtained. (solver termination condition: {0})".format(
                        results.solver.termination_condition.key
                    )
                )
            )
        else:
            self._process_results()

        return self.get_results()

    def set_model_parameters(self, **kwargs):
        """Set model parameters in kwargs to their respective values."""
        for kw_key, kw_value in kwargs.items():
            logging.info(
                "Optimizer: Setting {param} to {value}".format(
                    param=kw_key, value=kw_value
                )
            )
            setattr(self.model, kw_key, kw_value)
            setattr(self.model, kw_key, kw_value)
            setattr(self.model, kw_key, kw_value)
            setattr(self.model, kw_key, kw_value)



# This is for setting the window parameters like the initial size. Goes before any other import statements.
from kivy.config import Config

Config.set('graphics', 'height', '720')
Config.set('graphics', 'width', '1280')
Config.set('graphics', 'minimum_height', '720')
Config.set('graphics', 'minimum_width', '1280')
Config.set('graphics', 'resizable', '1')
Config.set('kivy', 'desktop', 1)

import os

from kivy.app import App
from kivy.app import Builder
from kivy.uix.widget import Widget
from kivy.uix.boxlayout import BoxLayout
from kivy.core.text import LabelBase

from es_gui.proving_grounds.help_carousel import HelpCarouselModalView

Builder.load_file(os.path.join('es_gui', 'resources', 'widgets', 'common.kv'))

LabelBase.register(name='Exo 2',
                   fn_regular=os.path.join('es_gui', 'resources', 'fonts', 'Exo_2', 'Exo2-Regular.ttf'),
                   fn_bold=os.path.join('es_gui', 'resources', 'fonts', 'Exo_2', 'Exo2-Bold.ttf'),
                   fn_italic=os.path.join('es_gui', 'resources', 'fonts', 'Exo_2', 'Exo2-Italic.ttf'))

LabelBase.register(name='Open Sans',
                   fn_regular=os.path.join('es_gui', 'resources', 'fonts', 'Open_Sans', 'OpenSans-Regular.ttf'),
                   fn_bold=os.path.join('es_gui', 'resources', 'fonts', 'Open_Sans', 'OpenSans-Bold.ttf'),
                   fn_italic=os.path.join('es_gui', 'resources', 'fonts', 'Open_Sans', 'OpenSans-Italic.ttf'))

LabelBase.register(name='Modern Pictograms',
                   fn_regular=os.path.join('es_gui', 'resources', 'fonts', 'modernpictograms', 'ModernPictograms.ttf'))


class Home(BoxLayout):
    def open_help(self):
        HelpCarouselModalView().open()


class HelpCarouselTestApp(App):
    def build(self):
        return Home()


if __name__ == '__main__':
    HelpCarouselTestApp().run()


from __future__ import absolute_import
import logging
import os
import json

import pandas as pd

from es_gui.tools.dms import DataManagementSystem
from es_gui.tools.btm.readutdata import *


class BtmDMS(DataManagementSystem):
    """
    A class for managing data for the behind-the-meter energy storage functions. Class methods for each type of file to be loaded are included, extending from the get_data() method of the superclass. Each of these methods uses get_data() to retrieve the relevant data and loads the file and adds it to the DMS if the data is not loaded. An optional class method for calling each of the individual data methods can be included to, e.g., form the necessary arguments and return the desired variables.

    :param home_path: A string indicating the relative path to where data is saved.
    """
    def __init__(self, home_path, **kwargs):
        DataManagementSystem.__init__(self, **kwargs)

        self.home_path = home_path
        self.delimiter = ' @ '  # delimiter used to split information in id_key
    
    def get_load_profile_data(self, path, month):
        """Retrieves commercial or residential load profile data."""
        logging.info('DMS: Loading load profile data')

        month = str(month)
        load_profile_key = self.delimiter.join([path, month])

        try:
            load_profile = self.get_data(load_profile_key)
        except KeyError:
            load_profile = read_load_profile(path, month)
            self.add_data(load_profile, load_profile_key)
        finally:
            return load_profile
    
    def get_pv_profile_data(self, path, month):
        """Retrieves PV profile data."""
        logging.info('DMS: Loading PV profile data')

        month = str(month)
        pv_profile_key = self.delimiter.join([path, month])

        try:
            pv_profile = self.get_data(pv_profile_key)
        except KeyError:
            pv_profile = read_pv_profile(path, month)
            self.add_data(pv_profile, pv_profile_key)
        finally:
            return pv_profile


from __future__ import absolute_import

import threading
import logging
from time import sleep
from functools import partial

from kivy.uix.screenmanager import Screen
from kivy.clock import Clock
from kivy.properties import ObjectProperty
from kivy.clock import mainthread
from kivy.uix.button import Button
from kivy.uix.popup import Popup
from kivy.animation import Animation
from kivy.app import App

from es_gui.tools.valuation.valuation_optimizer import ValuationOptimizer
from es_gui.apps.valuation.loaddatascreen import InputError


class ValuationScreen(Screen):
    """The home screen for the 'Valuation - Advanced' section."""
    def __init__(self, **kwargs):
        super(ValuationScreen, self).__init__(**kwargs)

        self.op = ValuationOptimizer()

        self.pw = ProgressWindow()
        self.pw.run_button.bind(on_release=self.execute_valuation)

    def on_enter(self):
        ab = self.manager.nav_bar
        ab.build_valuation_advanced_nav_bar()
        ab.set_title('Valuation - Advanced')

    def _generate_requests(self):
        data_screen = self.manager.get_screen('load_data')
        params_screen = self.manager.get_screen('set_parameters')

        iso_selected, market_formulation_selected, node_selected, year_selected, month_selected = data_screen.get_inputs()

        param_settings = params_screen.get_inputs()

        requests = {'iso': iso_selected,
                    'market type': market_formulation_selected,
                    'months': [(month_selected, year_selected)],
                    'node id': node_selected,
                    }

        if param_settings:
            requests['param set'] = param_settings

        return requests

    def execute_valuation(self, *args):
        try:
            requests = self._generate_requests()
        except InputError as e:
            logging.error('ValuationExecute: ' + str(e))

            popup = WarningPopup()
            popup.popup_text.text = str(e)
            popup.open()
        else:
            self.pw.title = 'Running optimization'
            self.pw.run_button.disabled = True

            solver_name = App.get_running_app().config.get('optimization', 'solver')
            handler = self.manager.get_screen('valuation_home').handler
            handler.solver_name = solver_name
            handler.process_requests(requests)

            self.pw.update_window('Finished!', 100, anim_duration=0)
            self.pw.title = 'Finished!'

            self.pw.view_results_button.bind(on_release=self.activate_view_results_button)
            self.pw.view_results_button.disabled = False

    def open_valuation_run_menu(self, *args):
        """
        Opens the popup/screen for executing a valuation optimization routine.
        """
        #if self.op.price_electricity is not None:
        if True:
            self.pw.open()
        else:
            # Optimizer LMP attribute is still its initialized None value.
            # TODO: Better way to check if data has been loaded?

            popup = WarningPopup()
            popup.popup_text.text = 'No data currently loaded. Please load data first!'
            popup.open()

    def save_run(self, name, iso, dtype, year, month, node, op=None):
        from datetime import datetime

        time_finished = datetime.now().strftime('%A, %B %d, %Y %H:%M:%S')

        results_dict = {}
        results_dict['name'] = name
        if op is None:
            results_dict['op'] = self.op
        else:
            results_dict['op'] = op
        results_dict['iso'] = iso
        results_dict['data_type'] = dtype
        results_dict['year'] = year
        results_dict['month'] = month
        results_dict['node'] = node
        results_dict['time'] = time_finished

        plot_screen = self.manager.get_screen('valuation_results_viewer')
        rv = plot_screen.run_selector.rv
        rv.data.append(results_dict)

    def execute_valuation_run(self, *args):
        """
        Executes the valuation run in a separate thread when possible.
        """
        self.pw.title = 'Running optimization'
        self.pw.run_button.disabled = True

        def _execute_valuation_run(*largs):
            self.pw.progress_label.text = ''

            # instantiate Pyomo model
            anim_duration = 0.3
            self.pw.update_window('Instantiating Pyomo model...\n', 10, anim_duration=anim_duration)
            self.op.instantiate_model()

            # populate Pyomo model
            anim_duration = 2.5  # this duration is empirically based
            self.pw.update_window('Populating Pyomo model...\n', 80, anim_duration=anim_duration)
            self.op.populate_model()

            # solve and process results
            anim_duration = 0.5
            self.pw.update_window('Solving model...\n', 90)
            sleep(0.5)  # hack to force pb to update before the solve clogs the main loop/thread

            @mainthread
            def solve_pyomo_model():
                # get the current optimization solver from settings
                solver_name = App.get_running_app().config.get('optimization', 'solver')
                self.op.solver = solver_name
                self.op.solve_model()

                # post-process solve results
                anim_duration = 0.5
                self.pw.update_window('Processing solve results...\n', 100, anim_duration=anim_duration)

                sleep(anim_duration)  # let the animation play out before "finish"?

                # finish
                anim_duration = 0
                self.pw.update_window('Finished!', 100, anim_duration=anim_duration)
                self.pw.title = 'Finished!'
                self.pw.view_results_button.bind(on_release=self.activate_view_results_button)
                self.pw.view_results_button.disabled = False

                # add to finished runs
                from random import sample
                from string import ascii_letters

                load_screen = self.manager.get_screen('load_data')
                iso = load_screen.iso_select.text
                year = load_screen.year_select.text
                month = load_screen.month_select.text
                dtype = load_screen.type_select.text
                node = load_screen.node_select.text

                # random string identifier for uniqueness
                name = '-'.join([node, year, month, ''.join(sample(ascii_letters, 5))])

                self.save_run(name, iso, dtype, year, month, node)

                # create new instance of Optimizer
                self.op = ValuationOptimizer()

            solve_pyomo_model()

        thread = threading.Thread(target=_execute_valuation_run, args=())
        thread.start()

    def activate_view_results_button(self, *args):
        """
        Enables the "View Results" button after the optimization routine is complete.
        """
        self.manager.nav_bar.go_to_screen('valuation_results_viewer')
        self.pw.dismiss()


class MyPopup(Popup):
    pass


class WarningPopup(MyPopup):
    pass


class ProgressWindow(MyPopup):
    def on_dismiss(self):
        if self.progress_bar.value == self.progress_bar.max:
            # Reset window if previous optimization has finished.
            self.reset_window()

    @mainthread
    def reset_window(self):
        """Resets the valuation run menu/screen to its initial state."""
        self.progress_label.text = 'Click "Run" to proceed.\n'
        self.progress_bar.value = 0
        self.title = 'Run optimization?'
        self.view_results_button.disabled = True
        self.run_button.disabled = False

    @mainthread
    def update_window(self, text, pb_val, anim_duration=0.5, *largs):
        """
        Updates the graphics of the valuation run menu/screen.

        :param text: The new text to be appended to the currently displayed text.
        :param pb_val: The value of the ProgressBar to be set.
        :param anim_duration: The duration of the ProgressBar animation.
        :param largs:
        :return:
        """
        pb_anim = Animation(value=pb_val, duration=anim_duration)
        Animation.cancel_all(self.progress_bar, 'value')

        self.progress_label.text += text
        pb_anim.start(self.progress_bar)



"""Market data processing tools."""
# from __future__ import absolute_import
# from xlrd.biffh import XLRDError

import pandas as pd
import numpy as np
import os
import calendar
import logging


def read_nodeid(fname, iso):
    """Read node id from an excel file and return as a list."""
    from xlrd import open_workbook

    wb = open_workbook(filename=fname)
    ws = wb.sheet_by_name(iso)
    nodeid = []
    for i in range(1, ws.nrows - 1):
        cell = ws.cell(i, 0)
        text = str(cell.value)
        text = text.replace(".0", "")
        nodeid += [text]
    return nodeid

def read_miso_da_lmp(fname, node_name):
    """
    Read MISO day-ahead LMP file and return the NumPy ndarray corresponding to the hourly LMP at node_name.

    :param fname: A string containing the path to the relevant day-ahead LMP file.
    :param node_name: A string containing the name of the pricing node of interest.
    :return: LMP: A NumPy ndarray containing the hourly LMP at node-name.
    """
    # parse fname for month and year values
    month = int(fname[-2:])
    year = int(fname[-6:-2])

    nday = {
        1: 31,
        2: 28,
        3: 31,
        4: 30,
        5: 31,
        6: 30,
        7: 31,
        8: 31,
        9: 30,
        10: 31,
        11: 30,
        12: 31,
    }
    if year % 4 == 0:
        nday[2] = 29

    LMP = np.empty([0])
    for day in range(1, nday[month] + 1):

        if (year <= 2014) or (year == 2015 and month <= 2):
            fname_ = fname + str(day).zfill(2) + "_da_lmp.csv"
        else:
            fname_ = fname + str(day).zfill(2) + "_da_exante_lmp.csv"

        df = pd.read_csv(fname_, skiprows=4, low_memory=False)
        # filter rows by node_name
        col1 = df.axes[1][0]
        col3 = df.axes[1][2]
        pnode_ix1 = df.index[df[col1] == node_name]
        df1 = df.iloc[pnode_ix1, :]

        # find LMP values
        pnode_ix2 = df1.index[df1[col3] == "LMP"]
        df2 = df.iloc[pnode_ix2, :]

        # filter Total LMP columns
        df3 = df2[df2.axes[1][3:27]]

        # convert to NumPy ndarray, ravel, and remove NaNs
        LMP_day = np.ravel(df3.astype("float").values)
        LMP_day = LMP_day[~np.isnan(LMP_day)]
        LMP = np.append(LMP, LMP_day)

    return LMP


def read_miso_reg_price(fname):
    """
    Read MISO historical ancillary services data file at fname and returns the NumPy ndarrays for regulation clearing prices.

    :param fname: A string containing the path to the relevant historical ancillary services data file.
    :return: RegMCP: NumPy ndarrays containing hourly regulation capacity/mileage clearing price values.
    """
    # parse fname for month and year values
    month = int(fname[-2:])
    year = int(fname[-6:-2])

    nday = {
        1: 31,
        2: 28,
        3: 31,
        4: 30,
        5: 31,
        6: 30,
        7: 31,
        8: 31,
        9: 30,
        10: 31,
        11: 30,
        12: 31,
    }
    if year % 4 == 0:
        nday[2] = 29

    RegMCP = np.empty([0])

    for day in range(1, nday[month] + 1):
        if (year <= 2014) or (year == 2015 and month <= 2):
            fname_ = fname + str(day).zfill(2) + "_asm_damcp.csv"
        else:
            fname_ = fname + str(day).zfill(2) + "_asm_exante_damcp.csv"

        # read in the .csv file
        df = pd.read_csv(fname_, skiprows=4, nrows=7, low_memory=False)

        # find SERREGMCP values
        col3 = df.axes[1][2]
        pnode_ix1 = df.index[df[col3] == "SERREGMCP"]
        df1 = df.iloc[pnode_ix1, :]
        df2 = df1[df1.axes[1][3:27]]

        # convert to NumPy ndarray, ravel, and remove NaNs
        RegMCP_day = np.ravel(df2.astype("float").values)
        RegMCP_day = RegMCP_day[~np.isnan(RegMCP_day)]
        RegMCP = np.append(RegMCP, RegMCP_day)

    return RegMCP

def read_miso_data(fpath, year, month, nodeid):
    """Read the daily MISO data files and returns the NumPy ndarrays for LMP and MCP.

    :param fpath: root of the MISO data folder
    :type fpath: str
    :param year: year of data
    :type year: int or str
    :param month: month of data
    :type month: int or str
    :param nodeid: pricing node ID
    :type nodeid: str
    :return: arrays of data specified
    :rtype: NumPy ndarrays
    """
    LMP = np.array([])
    RegMCP = np.array([])

    _, n_days_month = calendar.monthrange(int(year), int(month))

    for day in range(1, n_days_month + 1):
        # Read daily files.
        date_str = "{year}{month}{day}".format(
            year=year, month=str(month).zfill(2), day=str(day).zfill(2)
        )

        if (int(year) <= 2014) or (int(year) == 2015 and int(month) <= 2):
            lmp_fname = os.path.join(
                fpath,
                "LMP",
                str(year),
                str(month).zfill(2),
                "{prefix}_da_lmp.csv".format(prefix=date_str),
            )
            mcp_fname = os.path.join(
                fpath,
                "MCP",
                str(year),
                str(month).zfill(2),
                "{prefix}_asm_damcp.csv".format(prefix=date_str),
            )
        else:
            lmp_fname = os.path.join(
                fpath,
                "LMP",
                str(year),
                str(month).zfill(2),
                "{prefix}_da_exante_lmp.csv".format(prefix=date_str),
            )
            mcp_fname = os.path.join(
                fpath,
                "MCP",
                str(year),
                str(month).zfill(2),
                "{prefix}_asm_exante_damcp.csv".format(prefix=date_str),
            )

        # LMP file.
        try:
            df = pd.read_csv(lmp_fname, skiprows=4, low_memory=False)
        except FileNotFoundError:
            logging.warning("read_miso_data: LMP file missing, returning empty array.")
            break

        # Filter rows by node_name.
        col1 = df.axes[1][0]
        col3 = df.axes[1][2]
        pnode_ix1 = df.index[df[col1] == nodeid]
        df1 = df.iloc[pnode_ix1, :]

        # Find LMP values.
        pnode_ix2 = df1.index[df1[col3] == "LMP"]
        df2 = df.iloc[pnode_ix2, :]

        # Filter Total LMP columns.
        df3 = df2[df2.axes[1][3:27]]

        if len(df3) == 0:
            LMP = np.array([])
            logging.warning(
                "read_miso_data: A daily LMP file is missing required data, returning empty array."
            )
            break

        # Convert to NumPy ndarray, ravel, and remove NaNs.
        LMP_day = np.ravel(df3.astype("float").values)
        LMP_day = LMP_day[~np.isnan(LMP_day)]
        LMP = np.append(LMP, LMP_day)

        # MCP file.
        try:
            df = pd.read_csv(mcp_fname, skiprows=4, nrows=7, low_memory=False)
        except FileNotFoundError:
            RegMCP = np.array([])
            logging.warning("read_miso_data: MCP file missing, returning empty array.")
            break

        # Find SERREGMCP values.
        col3 = df.axes[1][2]
        pnode_ix1 = df.index[df[col3] == "SERREGMCP"]
        df1 = df.iloc[pnode_ix1, :]
        df2 = df1[df1.axes[1][3:27]]

        # convert to NumPy ndarray, ravel, and remove NaNs
        RegMCP_day = np.ravel(df2.astype("float").values)
        RegMCP_day = RegMCP_day[~np.isnan(RegMCP_day)]
        RegMCP = np.append(RegMCP, RegMCP_day)

    return LMP, RegMCP


from functools import partial
from collections import namedtuple, OrderedDict
from abc import ABCMeta, abstractmethod, abstractproperty

from kivy.graphics import Ellipse, Color, Rectangle, Line
from kivy.animation import Animation
from kivy.properties import NumericProperty, ListProperty
from kivy.uix.widget import Widget
from kivy.uix.label import Label
from kivy.clock import Clock
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.relativelayout import RelativeLayout
from kivy.uix.gridlayout import GridLayout
from kivy.uix.stencilview import StencilView


def format_dollar_string(value):
    """Formats a float representing USD with comma separators, dollar sign, cents precision, and proper negative sign placement."""
    if value >= 0:
        return_str = "${0:,.2f}".format(value)
    else:
        return_str = "-${0:,.2f}".format(-value)
    
    return return_str


class Chart(RelativeLayout):
    def __init__(self, do_animation=True, **kwargs):
        super(Chart, self).__init__(**kwargs)

        self._do_animation = do_animation
        self._is_drawn = False

    @property
    def do_animation(self):
        """Set to True to animate the chart when the draw_chart() method is called."""
        return self._do_animation

    @do_animation.setter
    def do_animation(self, value):
        self._do_animation = value
    
    @property
    def is_drawn(self):
        """Returns True if a chart has been drawn."""
        return self._is_drawn
    
    @is_drawn.setter
    def is_drawn(self, value):
        self._is_drawn = value

    @abstractmethod
    def draw_chart(self, data):
        pass

    def export_chart(self, data, fname):
        self.draw_chart(data)
        self.export_to_png(fname)


class Bar(Widget):
    height_final = NumericProperty

    def __init__(self, rgba=None, height_final=None, info=None, **kwargs):
        """
        A widget for bars in bar charts.

        :param rgba: Length-4 iterable representing R, G, B, Alpha values in [0, 1].
        :param height_final: Float representing the value of the bar.
        :param info: BarChartEntry, a named tuple with fields ('name', 'rgba', 'value')
        :param kwargs:
        """
        super(Bar, self).__init__(**kwargs)

        self.size_hint_x = None
        self.size_hint_y = None

        self.info = info

        self.color = rgba
        self.height_final = height_final

        with self.canvas.after:
            # draws the bar as a rectangle graphic
            Color(*rgba)
            self.bar = Rectangle(pos=self.pos, size=self.size)

        self.bind(pos=self._update_bar, size=self._update_bar)

    def _update_bar(self, instance, value):
        """Updates Bar graphic size and position whenever widget size or position changes."""
        self.bar.pos = instance.pos
        self.bar.size = instance.size


class StackedBarChart(Chart):
    def __init__(self, y_axis_format=None, bar_spacing=50, legend_width=170, x_padding=50, y_padding=50, **kwargs):
        """
        A vertically-orientated stacked bar chart. Call draw-chart() with bar_data to draw the chart.

        :param y_axis_format: str; format string for the y-axis labels. Namely for determining precision/separators/etc.
        :param bar_spacing: int (px) representing the space between adjacent bar groups.
        :param legend_width: int (px) representing the width of the legend. Increase if the labels do not render completely.
        :param x_padding: int (px) representing the space between the left and right edges of the chart and chart elements.
        :param y_padding: int (px) representing the space between the top and bottom edges of the chart and chart elements.
        """
        super(StackedBarChart, self).__init__(**kwargs)

        # sets default y-axis label format to two decimal place precision (fixed point)
        if not y_axis_format:
            self.y_axis_format = '{0:.2f}'
        else:
            self.y_axis_format = y_axis_format

        self.legend_width = legend_width
        self.bar_spacing = bar_spacing
        self.x_padding = x_padding
        self.y_padding = y_padding

        self.max_height = None
        self.bar_width = None

        self.max_bar = None
        self.min_bar = None

        self.bar_groups = OrderedDict()
        self.legend = None

    def generate_bars(self, bar_data):
        """
        Computes the size and positions of the bars using bar_data. Adds the bars and axis labels to StackedBarChart. Only supports data values >= 0.

        :param bar_data: OrderedDict where key: name and value: list of lists in format ['category', 'color (rgba iterable)', 'data value'].
        """
        # create a namedtuple to keep track of the information needed to describe each bar stack component
        BarChartEntry = namedtuple('BarChartEntry', ['category', 'rgba', 'value'])
        bars = OrderedDict()

        # form dictionary with key: bar stack name and value: list of BarChartEntry namedtuples corresponding to stack components
        for bar_set_name in bar_data.keys():
            bars[bar_set_name] = [BarChartEntry._make(x) for x in bar_data[bar_set_name]]

            # catch any negative-valued data and raise exception
            if any([bar_component.value < 0 for bar_component in bars[bar_set_name]]):
                raise (ValueError('StackedBarChart does not support negative values!'))

        # determine the tallest and shortest stacks by adding up their values
        self.max_bar = max(bars.items(), key=lambda bar: sum([x.value for x in bar[1]]))
        self.min_bar = min(bars.items(), key=lambda bar: sum([x.value for x in bar[1]]))

        # determine the maximum and minimum stack values
        max_value = sum([x.value for x in self.max_bar[1]])
        min_value = sum([x.value for x in self.min_bar[1]])

        # compute origin coordinates, maximum bar height, and bar width
        x0 = self.legend_width + self.x_padding  # x-coordinate of leftmost bar
        y0 = self.y_padding  # y-coordinate of bar bottom
        self.max_height = self.height - 2*self.y_padding  # y-coordinate of top of tallest bar

        # calculate the total amount of width available for allocating to bars and distributing it evenly
        self.bar_width = (self.width - x0 - self.x_padding - self.bar_spacing*(len(bars) - 1))/len(bars)

        # compute bar height:value ratio
        dydv = self.max_height/(max_value - min(0, min_value))  # ensure that y=0 is included in chart

        # iterate over each bar stack
        for ix, bar_list in enumerate(bars.items(), start=0):
            # unpack tuple
            name = bar_list[0]
            bars = bar_list[1]

            # determine bar position based on index
            bar_pos = (x0 + ix * (self.bar_width + self.bar_spacing), y0)

            # create x-axis label using name
            bar_label = Label(pos=(bar_pos[0] - self.width/2 + self.bar_width/2, self.y_padding/2 - self.height/2),
                              text=name, color=[0, 0, 0, 1])
            self.add_widget(bar_label)

            # initialize stack position and bar stack list
            initial_pos = (bar_pos[0], y0)
            self.bar_groups[name] = []

            # iterate over each component of the stack
            for bar in bars:
                # calculate height based on value
                bar_height = bar.value*dydv

                # generate bar widget and add to appropriate stack
                bar_widget = Bar(size=(self.bar_width, 0), pos=initial_pos, rgba=bar.rgba, height_final=bar_height)
                self.bar_groups[name].append(bar_widget)

                # calculate new starting y-coordinate for next stack element
                initial_pos = (initial_pos[0], initial_pos[1]+bar_height)

        # create y-axis label for maximum value in chart
        max_label = Label(pos=(0.75*x0 - self.width/2, self.max_height + self.y_padding - self.height/2),
                              text=self.y_axis_format.format(max_value), halign='right', color=[0, 0, 0, 1])
        self.add_widget(max_label)

        # line across chart to indicate maximum value
        with self.canvas:
            Color(0, 0, 0, 1)
            Line(points=[x0, self.max_height + self.y_padding,
                         self.width - self.x_padding, self.max_height + self.y_padding],
                 width=1,)

        # create y-axis label for y=0
        zero_label = Label(pos=(0.75*x0 - self.width/2, self.max_height - max_value*dydv + self.y_padding - self.height/2),
                           text='0', halign='right', color=[0, 0, 0, 1])
        self.add_widget(zero_label)

    def draw_chart(self, bar_data):
        """
        Draws the StackedBarChart.

        :param bar_data: OrderedDict where key: name and value: list of lists in format ['category', 'color (rgba iterable)', 'data value'].
        """
        # clear all widgets from the chart
        while len(self.children) > 0:
            for widget in self.children:
                self.remove_widget(widget)

        with self.canvas.before:
            self.canvas.clear()
            Color(1, 1, 1, 1)
            Rectangle(pos=self.pos, size=self.size)

        # generate bar widgets
        self.generate_bars(bar_data)

        if not self._do_animation:
            t_anim = 0  # the length [s] of the growing bar element animation
        else:
            t_anim = 0.5
        n_stack_comp = max([len(x) for x in bar_data.values()])  # the number of elements in each group

        def _anim_bar(bar, height, *args):
            anim = Animation(size=(bar.size[0], height), duration=t_anim, t='out_circ')
            anim.start(bar)

        # iterate over the bar groups
        for name in bar_data.keys():

            # iterate over the bar group elements
            for ix, bar in enumerate(self.bar_groups[name], start=0):
                # add the bar to the chart and schedule its animation
                self.add_widget(bar)
                Clock.schedule_once(partial(_anim_bar, bar, bar.height_final), ix*t_anim)

        # generate legend
        leg_pos = (0, 0)
        leg_size = (self.legend_width, self.height)

        self.legend = ChartLegend(leg_pos, size=leg_size, opacity=0)
        self.add_widget(self.legend)

        # form the legend input [[name, rgba] for each entry]
        legend_data = [[x[0], x[1]] for x in bar_data[list(bar_data.keys())[0]]]

        self.legend.gen_legend(legend_data)

        def _anim_legend(legend, *args):
            if self._do_animation:
                t_anim_legend = 1.0
            else:
                t_anim_legend = 0

            anim = Animation(opacity=1, duration=t_anim_legend, t='linear')
            anim.start(legend)

        # animate the legend opacity after all bars have been built
        Clock.schedule_once(partial(_anim_legend, self.legend), t_anim*n_stack_comp)

        self.is_drawn = True


class MultisetBarChart(StackedBarChart):
    def __init__(self, y_axis_format=None, bar_spacing=50, legend_width=160, x_padding=50, y_padding=50, **kwargs):
        """
        A vertically-orientated multi-set bar chart. Call draw-chart() with bar_data to draw the chart. Subclassed from StackedBarChart.

        :param y_axis_format: str; format string for the y-axis labels. Namely for determining precision/separators/etc.
        :param bar_spacing: int (px) representing the space between adjacent bar groups.
        :param legend_width: int (px) representing the width of the legend. Increase if the labels do not render completely.
        :param x_padding: int (px) representing the space between the left and right edges of the chart and chart elements.
        :param y_padding: int (px) representing the space between the top and bottom edges of the chart and chart elements.
        """
        super(MultisetBarChart, self).__init__(**kwargs)

        # sets default y-axis label format to two decimal place precision (fixed point)
        if not y_axis_format:
            self.y_axis_format = '{0:.2f}'
        else:
            self.y_axis_format = y_axis_format

        self.bar_spacing = bar_spacing
        self.legend_width = legend_width
        self.x_padding = x_padding
        self.y_padding = y_padding

        self.max_height = None
        self.bar_group_width = None
        self.bar_width = None

        self.max_bar = None
        self.min_bar = None

        self.bar_groups = OrderedDict()
        self.legend = None

    def generate_bars(self, bar_data):
        """
        Computes the size and positions of the bars using bar_data. Adds the bars and axis labels to MultisetBarChart.

        :param bar_data: OrderedDict where key: name and value: list of lists in format ['category', 'color (rgba iterable)', 'data value'].
        """
        # create a namedtuple to keep track of the information needed to describe each bar group component
        BarChartEntry = namedtuple('BarChartEntry', ['category', 'rgba', 'value'])
        bars = OrderedDict()

        # form dictionary with key: bar stack name and value: list of BarChartEntry tuples corresponding to set components
        for bar_set_name in bar_data.keys():
            bars[bar_set_name] = [BarChartEntry._make(x) for x in bar_data[bar_set_name]]
        
        # max value by category
        all_bar_components = [bar for bar_set in bars.values() for bar in bar_set]
        categories = {bar_component.category for bar_component in all_bar_components}

        max_bar_entries = {}

        for category in categories:
            category_components = []

            for bar_component in all_bar_components:
                if bar_component.category == category:
                    category_components.append(bar_component)

            max_bar_entries[category] = max(category_components, key=lambda x: abs(x.value))

        # determine the tallest and shortest bars and their respective values
        self.max_bar = max([bar_entry for bar_set in bars.values() for bar_entry in bar_set], key=lambda x: x.value)
        self.min_bar = min([bar_entry for bar_set in bars.values() for bar_entry in bar_set], key=lambda x: x.value)

        max_value = self.max_bar.value
        min_value = self.min_bar.value

        # compute origin coordinates, maximum bar height, and bar width
        x0 = self.legend_width + self.x_padding  # x-coordinate of leftmost bar

        self.max_height = self.height - 2 * self.y_padding  # y-coordinate of top of tallest bar

        # calculate the total amount of width available for allocating to bar groups and distributing it evenly
        self.bar_group_width = (self.width - x0 - self.x_padding - self.bar_spacing * (len(bars) - 1)) / len(bars)

        # calculate width of bars in each in group, assuming no spacing among them
        self.bar_width = self.bar_group_width /max([len(x) for x in bars.values()])

        # compute bar height:value ratio
        try:
            dydv = self.max_height/(max(0, max_value) - min(0, min_value))
        except ZeroDivisionError:
            dydv = 1

        # if negative values: determine coordinate of y=0
        if min_value < 0:
            y0 = self.y_padding + abs(min_value) * dydv
        else:
            y0 = self.y_padding
        
        if max_value < 0:
            y0 = self.max_height + self.y_padding

        # iterate over each bar group
        for ix, bar_list in enumerate(bars.items(), start=0):
            # unpack tuple
            name = bar_list[0]
            bars = bar_list[1]

            # determine bar group position based on index
            bar_pos = (x0 + ix * (self.bar_group_width + self.bar_spacing), y0)

            # create x-axis label using name
            bar_label = Label(pos=(bar_pos[0] - self.width/2 + self.bar_group_width/2, self.y_padding/2 - self.height/2),
                              text=name, color=[0, 0, 0, 1])
            self.add_widget(bar_label)

            # initialize group position and bar group list
            initial_pos = (bar_pos[0], y0)
            self.bar_groups[name] = []

            # iterate over each bar in the group
            for bar in bars:
                # calculate height based on value
                bar_height = bar.value*dydv

                # generate bar widget and add to appropriate group
                bar_widget = Bar(size=(self.bar_width, 0), pos=initial_pos, rgba=bar.rgba, height_final=bar_height)
                self.bar_groups[name].append(bar_widget)

                # calculate new starting x-coordinate for next bar
                initial_pos = (initial_pos[0] + self.bar_width, initial_pos[1])

        # create y-axis label for maximum value in chart
        if max_value > 0:
            max_label = Label(pos=(0.75*x0 - self.width/2, self.max_height + self.y_padding - self.height/2),
                                text=self.y_axis_format.format(max_value), halign='right', color=[0, 0, 0, 1])
            self.add_widget(max_label)

            # line across chart to indicate maximum value
            with self.canvas:
                Color(0, 0, 0, 1)
                Line(points=[x0, self.max_height + self.y_padding,
                            self.width - self.x_padding, self.max_height + self.y_padding],
                    width=1,)

            # create y-axis label for y=0
            zero_label = Label(pos=(0.75*x0 - self.width/2, self.max_height - max_value*dydv + self.y_padding - self.height/2),
                            text='0', halign='right', color=[0, 0, 0, 1])
        else:
            with self.canvas:
                # line to indicate zero value
                Color(0, 0, 0, 1)
                Line(points=[x0, self.max_height + self.y_padding,
                            self.width - self.x_padding, self.max_height + self.y_padding],
                    width=1,)

            zero_label = Label(pos=(0.75*x0 - self.width/2, self.max_height + self.y_padding - self.height/2),
                           text='0', halign='right',color=[0, 0, 0, 1])  
                           
        self.add_widget(zero_label)

        # label y=0 if minimum value < 0
        if min_value < 0:
            # fix y_axis_format specification for proper $ representation for negative values
            if self.y_axis_format[0] == '$':
                label_text = '-' + self.y_axis_format.format(-min_value)
            else:
                label_text = self.y_axis_format.format(min_value)

            min_label = Label(pos=(0.75*x0 - self.width/2, self.y_padding - self.height/2),
                              text=label_text, halign='right', color=[0, 0, 0, 1])
            self.add_widget(min_label)

            # line across chart to indicate minimum value
            with self.canvas:
                Color(0, 0, 0, 1)
                Line(points=[x0, self.y_padding,
                             self.width - self.x_padding, self.y_padding],
                     width=1,)
        
        # label the maximum value for each multibar set
        for category in categories:
            category_max_bar = max_bar_entries[category]

            category_max_value = category_max_bar.value

            if category_max_value != max_value and category_max_value != min_value:
                label_text = self.y_axis_format.format(category_max_value)

                if category_max_value < 0:
                    if self.y_axis_format[0] == '$':
                        label_text = '-' + self.y_axis_format.format(-category_max_value)

                if max_value > 0:
                    component_max_label = Label(pos=(0.75*x0 - self.width/2, self.max_height - (max_value - category_max_value)*dydv + self.y_padding - self.height/2),
                                        text=label_text, halign='right', color=category_max_bar.rgba)
                    self.add_widget(component_max_label)

                    with self.canvas:
                        Color(*category_max_bar.rgba)
                        Line(points=[x0, self.max_height + self.y_padding - (max_value - category_max_value)*dydv,
                                    self.width - self.x_padding, self.max_height + self.y_padding - (max_value - category_max_value)*dydv],
                            width=1,)
                else:
                    component_max_label = Label(pos=(0.75*x0 - self.width/2, self.max_height + category_max_value*dydv + self.y_padding - self.height/2),
                                        text=label_text, halign='right', color=category_max_bar.rgba)
                    self.add_widget(component_max_label)

                    with self.canvas:
                        Color(*category_max_bar.rgba)
                        Line(points=[x0, self.max_height + self.y_padding + category_max_value*dydv,
                                    self.width - self.x_padding, self.max_height + self.y_padding + category_max_value*dydv],
                            width=1,)


class BarChart(Chart):
    def __init__(self, y_axis_format=None, bar_spacing=50, x_padding=25, y_padding=40, **kwargs):
        """
        A vertically-orientated bar chart. Call draw_chart() with bar_data to draw the chart.

        :param y_axis_format: str; format string for the y-axis labels. Namely for determining precision/separators/etc.
        :param bar_spacing: int (px) representing the space between adjacent bars.
        :param x_padding: int (px) representing the space between the left and right edges of the chart and chart elements.
        :param y_padding: int (px) representing the space between the top and bottom edges of the chart and chart elements.
        """
        super(BarChart, self).__init__(**kwargs)

        # sets default y-axis label format to two decimal place precision (fixed point)
        if not y_axis_format:
            self.y_axis_format = '{0:.2f}'
        else:
            self.y_axis_format = y_axis_format

        self.bar_spacing = bar_spacing
        self.x_padding = x_padding
        self.y_padding = y_padding

        self.max_height = None
        self.bar_width = None

        self.bars = []
        self.max_bar = None
        self.min_bar = None

    def generate_bars(self, bar_data):
        """
        Computes the size and positions of the bars using bar_data. Adds the bars and axis labels to BarChart.

        :param bar_data: List of lists in format ['name', 'color (rgba iterable)', 'data value'].
        """
        # create a namedtuple to keep track of the information needed to describe each bar
        BarChartEntry = namedtuple('BarChartEntry', ['name', 'rgba', 'value'])
        bars = [BarChartEntry._make(x) for x in bar_data]

        # determine the tallest and shortest bars
        self.max_bar = max(bars, key=lambda x: x.value)
        self.min_bar = min(bars, key=lambda x: x.value)

        max_value = self.max_bar.value  # the largest value among all prospective bars
        min_value = self.min_bar.value  # the smallest value among all prospective bars

        # compute origin coordinates, maximum bar height, and bar width
        x0 = 5*self.x_padding

        self.max_height = self.height - 2*self.y_padding

        # calculate the total amount of width available for allocating to bars and distributing it evenly
        self.bar_width = (self.width - x0 - self.x_padding - self.bar_spacing*(len(bars) - 1))/len(bars)

        # compute bar height:value ratio
        try:
            dydv = self.max_height/(max(0, max_value) - min(0, min_value))  # ensure that y=0 is included in chart
        except ZeroDivisionError:
            dydv = 1

        if min_value < 0:
            y0 = self.y_padding + abs(min_value)*dydv
        else:
            y0 = self.y_padding
        
        if max_value < 0:
            y0 = self.max_height + self.y_padding

        self.bars = []

        # iterate over each bar
        for ix, bar in enumerate(bars, start=0):
            # generate bars, labels, and bar heights
            bar_pos = (x0 + ix*(self.bar_width + self.bar_spacing), y0)

            bar_height = bar.value * dydv

            bar_widget = Bar(size=(self.bar_width, 0), pos=bar_pos, rgba=bar.rgba, height_final=bar_height, info=bar)
            bar_label = Label(pos=(bar_pos[0] - self.width/2 + self.bar_width/2, self.y_padding/2 - self.height/2),
                              text=bar.name, color=[0, 0, 0, 1])

            self.bars.append(bar_widget)
            self.add_widget(bar_widget)
            self.add_widget(bar_label)

        # create y-axis label for maximum value in chart
        if max_value > 0:
            max_label = Label(pos=(0.5*x0 - self.width/2, self.max_height + self.y_padding - self.height/2),
                            text=self.y_axis_format.format(max_value), halign='right', color=[0, 0, 0, 1])
            self.add_widget(max_label)

            # line across chart to indicate maximum value
            with self.canvas:
                Color(0, 0, 0, 1)
                Line(points=[x0, self.max_height + self.y_padding,
                            self.width - self.x_padding, self.max_height + self.y_padding],
                    width=1,)

            zero_label = Label(pos=(0.5*x0 - self.width/2, self.max_height - max_value*dydv + self.y_padding - self.height/2),
                           text='0', halign='right',color=[0, 0, 0, 1])
        else:
            with self.canvas:
                # line to indicate zero value
                Color(0, 0, 0, 1)
                Line(points=[x0, self.max_height + self.y_padding,
                            self.width - self.x_padding, self.max_height + self.y_padding],
                    width=1,)

            zero_label = Label(pos=(0.5*x0 - self.width/2, self.max_height + self.y_padding - self.height/2),
                           text='0', halign='right',color=[0, 0, 0, 1])        

        self.add_widget(zero_label)

        # label y=0 if minimum value < 0
        if min_value < 0:
            # fix y_axis_format specification for proper $ representation for negative values
            if self.y_axis_format[0] == '$':
                label_text = '-' + self.y_axis_format.format(-min_value)
            else:
                label_text = self.y_axis_format.format(min_value)

            min_label = Label(pos=(0.5 * x0 - self.width / 2, self.y_padding - self.height / 2),
                              text=label_text, halign='right', color=[0, 0, 0, 1])

            self.add_widget(min_label)

            # line across chart to indicate minimum value
            with self.canvas:
                Color(0, 0, 0, 1)
                Line(points=[x0, self.y_padding,
                             self.width - self.x_padding, self.y_padding],
                     width=1,)

    def draw_chart(self, bar_data):
        """
        Draws the BarChart.

        :param bar_data: List of lists in format ['name', 'color (rgba iterable)', 'data value'].
        """
        # clear all widgets from the BarChart
        while len(self.children) > 0:
            for widget in self.children:
                self.remove_widget(widget)

        with self.canvas.before:
            self.canvas.clear()
            Color(1, 1, 1, 1)
            Rectangle(size=self.size, pos=self.pos)

        self.generate_bars(bar_data)

        def _anim_bar(bar, height, *args):
            anim = Animation(size=(bar.size[0], height), duration=1.0, t='out_back')
            anim.start(bar)

        # animate the bars
        for ix, bar in enumerate(self.bars, start=0):
            if isinstance(bar, Bar):
                if self._do_animation:
                    Clock.schedule_once(partial(_anim_bar, bar, bar.height_final), ix*0.05)
                else:
                    bar.height = bar.height_final
        
        self.is_drawn = True


class DonutChart(Chart):
    def __init__(self, **kwargs):
        """
        A donut chart with a legend. Call draw_chart() with pie_data to draw the chart.
        """
        super(DonutChart, self).__init__(**kwargs)

        self.pie = None
        self.legend = None

    def draw_chart(self, pie_data):
        """
        Generates donut chart using pie_data and draws it.

        :param pie_data: List of lists in format ['name', 'color (rgba iterable)', 'data value']
        """
        # clear all widgets
        while len(self.children) > 0:
            for widget in self.children:
                self.remove_widget(widget)

        with self.canvas.before:
            self.canvas.clear()
            Color(1, 1, 1, 1)
            Rectangle(size=self.size, pos=self.pos)

        DonutChartEntry = namedtuple('DonutChartEntry', ['name', 'rgba', 'value'])
        slice_data = [DonutChartEntry._make(x) for x in pie_data]

        # compute donut size and pos
        pie_d = min(self.size[1], self.size[0]/2)
        pie_size = [pie_d, pie_d]
        pie_pos = [0, (self.size[1] - pie_d)/2]

        # generate pie and its slices
        self.pie = Pie(pos=pie_pos, size=pie_size, is_donut=True)
        self.add_widget(self.pie)
        self.pie.gen_slices(pie_data)

        # animate slices
        if self._do_animation:
            t_anim = 0.16
            t_anim_slice = 1.0
            t_anim_legend = 1.0
        else:
            t_anim = 0
            t_anim_slice = 0
            t_anim_legend = 0

        n_slices = len(pie_data)

        def _anim_slice(slice, *args):
            anim = Animation(opacity=1, duration=t_anim_slice, t='out_circ')
            anim.start(slice)

        for ix, slice in enumerate(self.pie.slices, start=0):
            Clock.schedule_once(partial(_anim_slice, slice), t_anim * ix)

        # generate legend
        leg_pos = [pie_pos[0] + pie_d, 0]
        leg_size = [pie_d, self.height]

        self.legend = ChartLegend(position=leg_pos, size=leg_size, opacity=0)
        #self.legend.padding = (0, self.height/4)
        self.add_widget(self.legend)

        val_total = float(sum([x.value if x.value > 0 else 0 for x in slice_data]))
        legend_data = []

        for entry in slice_data:
            legend_text = '{0:.2f}% {1}'.format(entry.value/val_total*100, entry.name)
            legend_data.append([legend_text, entry.rgba])

        self.legend.gen_legend(legend_data)

        # animate legend
        def _anim_legend(legend, *args):
            anim = Animation(opacity=1, duration=t_anim_legend, t='out_circ')
            anim.start(legend)

        Clock.schedule_once(partial(_anim_legend, self.legend), t_anim * n_slices)

        self.is_drawn = True

        # # generate legend
        # hole_proportion = 0.33
        # pie_radius = pie_d
        # hole_radius = pie_radius*(1 - hole_proportion)
        #
        # from math import sqrt
        #
        # leg_pos = [pie_pos[0] + hole_proportion*pie_radius, pie_pos[1] + hole_proportion*pie_radius]
        # leg_size = [2*hole_radius/sqrt(2), 2*hole_radius/sqrt(2)]
        #
        # self.legend = ChartLegend(leg_pos, leg_size, opacity=0)
        # with self.legend.canvas.before:
        #     Rectangle(pos=self.legend.pos, size=self.legend.size, color=Color(1, 0, 0, 0.1))
        # self.add_widget(self.legend)
        # self.legend.gen_legend(pie_data)
        #
        # # animate legend
        # def _anim_legend(legend, *args):
        #     anim = Animation(opacity=1, duration=1.0, t='out_circ')
        #     anim.start(legend)
        #
        # Clock.schedule_once(partial(_anim_legend, self.legend), t_anim * n_slices)


class PieChart(Chart):
    def __init__(self, **kwargs):
        """
        A pie chart with a legend. Call draw_chart() with pie_data to draw the chart.
        """
        super(PieChart, self).__init__(**kwargs)

        self.pie = None
        self.legend = None

    def draw_chart(self, pie_data, is_donut=False):
        """
        Generates pie chart and legend using pie_data and draws them.

        :param pie_data: List of lists in format ['name', 'color (rgba iterable)', 'data value']
        """
        # clear all widgets
        while len(self.children) > 0:
            for widget in self.children:
                self.remove_widget(widget)

        with self.canvas.before:
            self.canvas.clear()
            Color(1, 1, 1, 1)
            Rectangle(size=self.size, pos=self.pos)

        PieChartEntry = namedtuple('PieChartEntry', ['name', 'rgba', 'value'])
        slice_data = [PieChartEntry._make(x) for x in pie_data]

        # compute pie size and pos
        pie_d = min(self.size[1], self.size[0]/2)
        pie_size = [pie_d, pie_d]
        pie_pos = [0, (self.size[1] - pie_d)/2]

        # generate pie and its slices
        self.pie = Pie(pos=pie_pos, size=pie_size, is_donut=is_donut)
        self.add_widget(self.pie)
        self.pie.gen_slices(pie_data)

        # animate slices
        n_slices = len(self.pie.slices)
        if self._do_animation:
            t_anim = 0.16
            t_anim_slice = 1.0
            t_anim_legend = 1.0
        else:
            t_anim = 0
            t_anim_slice = 0
            t_anim_legend = 0

        def _anim_slice(slice, *args):
            anim = Animation(opacity=1, duration=t_anim_slice, t='out_circ')
            anim.start(slice)

        for ix, slice in enumerate(self.pie.slices, start=0):
            Clock.schedule_once(partial(_anim_slice, slice), t_anim*ix)

        # generate legend
        leg_pos = [pie_pos[0] + pie_d, 0]
        leg_size = [pie_d, self.height]

        self.legend = ChartLegend(leg_pos, size=leg_size, opacity=0)
        self.add_widget(self.legend)

        val_total = float(sum([x.value for x in slice_data]))
        legend_data = []

        for entry in slice_data:
            legend_text = '{0:.2f}% {1}'.format(entry.value/val_total*100, entry.name)
            legend_data.append([legend_text, entry.rgba])

        self.legend.gen_legend(legend_data)

        # animate legend
        def _anim_legend(legend, *args):
            anim = Animation(opacity=1, duration=t_anim_legend, t='out_circ')
            anim.start(legend)
        Clock.schedule_once(partial(_anim_legend, self.legend), t_anim*n_slices)

        self.is_drawn = True


class Pie(Widget):
    slices = ListProperty()

    def __init__(self, is_donut=False, separator_width=1.5, **kwargs):
        """
        A widget whose children are its constituting pie chart slices.

        :param is_donut: If True, cuts out a donut hole to transform it into a donut chart.
        :param separator_width: Width of line separating adjacent pie slices (px).
        """
        super(Pie, self).__init__(**kwargs)

        self.is_donut = is_donut
        self.separator_width = separator_width

    def gen_slices(self, pie_data):
        PieChartEntry = namedtuple('PieChartEntry', ['name', 'rgba', 'value'])
        slice_data = [PieChartEntry._make(x) for x in pie_data]
        slice_data.sort(key=lambda x: x.value, reverse=True)

        val_total = float(sum([x.value if x.value > 0 else 0 for x in slice_data]))

        self.slices = []

        total_arc = 360
        angle_start_pos = 0
        angle_start_neg = 0

        if self.is_donut:
            hole_proportion = 0.33  # "radius" of outer layer as fraction of given pie size
            pie_radius = self.size[0]/2
            hole_radius = pie_radius*(1 - hole_proportion)

            if any([entry.value < 0 for entry in slice_data]):
                # positive layer takes inner ring, negative takes original position
                positive_pos = (self.pos[0] + hole_proportion*pie_radius, self.pos[1] + hole_proportion*pie_radius)
                positive_radius = hole_radius
                negative_pos = self.pos
                negative_radius = pie_radius
            else:
                # positive layer only
                positive_pos = self.pos
                positive_radius = pie_radius

            for ix, slice_entry in enumerate(slice_data, start=0):
                # draw outer layer, negative values
                percentage = abs(slice_entry.value)/val_total
                color = slice_entry.rgba

                if slice_entry.value < 0:
                    angle_end_neg = angle_start_neg + percentage*total_arc
                    new_slice = PieSlice(pos=negative_pos,
                                         size=(negative_radius*2, negative_radius*2),
                                         angle_start=angle_start_neg, angle_end=angle_end_neg,
                                         color=color, name=slice_entry.name,
                                         separator_width=self.separator_width,
                                         opacity=0)
                    angle_start_neg = angle_end_neg

                    self.slices.append(new_slice)
                    self.add_widget(self.slices[ix])
                else:
                    break

            # cut outer donut hole
            with self.canvas:
                Color(1, 1, 1, 1)
                Ellipse(pos=positive_pos,
                        size=(2*positive_radius, 2*positive_radius), )

            for iy, slice_entry in enumerate(slice_data[ix:], start=ix):
                # draw inner layer, positive values
                percentage = slice_entry.value/val_total
                color = slice_entry.rgba

                if slice_entry.value > 0:
                    angle_end_pos = angle_start_pos + percentage*total_arc
                    new_slice = PieSlice(pos=positive_pos,
                                         size=(positive_radius*2, positive_radius*2),
                                         angle_start=angle_start_pos, angle_end=angle_end_pos,
                                         color=color, name=slice_entry.name,
                                         separator_width=self.separator_width,
                                         opacity=0)
                    angle_start_pos = angle_end_pos

                    self.slices.append(new_slice)
                    self.add_widget(self.slices[iy])

            with self.canvas:
                Color((1, 1, 1, 1))
                if any([entry.value < 0 for entry in slice_data]):
                    # inner hole
                    Ellipse(pos=(self.pos[0] + 2*hole_proportion*pie_radius, self.pos[1] + 2*hole_proportion*pie_radius),
                            size=(hole_radius, hole_radius),)

                    # pos/neg layer separator
                    Line(ellipse=(self.pos[0] + hole_proportion*pie_radius, self.pos[1] + hole_proportion*pie_radius,
                                  hole_radius*2, hole_radius*2), width=self.separator_width)
                else:
                    # inner hole
                    Ellipse(pos=(self.pos[0] + hole_proportion*pie_radius, self.pos[1] + hole_proportion*pie_radius),
                            size=(2*hole_radius, 2*hole_radius),)

            # label sectors
            # angle_start = 0
            # from math import cos, sin, pi
            #
            # for ix, slice_entry in enumerate(slice_data, start=0):
            #     percentage = slice_entry.value/val_total
            #     angle_end = angle_start + percentage*total_arc
            #     angle_midpoint = (angle_start + angle_end)/2
            #
            #     pct_label_pos = (pie_center[0] + 0.75*hole_radius*sin(angle_midpoint*pi/180),
            #                      pie_center[1] + 0.75*hole_radius*cos(angle_midpoint*pi/180))
            #
            #     pct_label = Label(text=str('%.2f' % (percentage*100) + '%'), color=(0, 0, 0, 1),
            #                       pos=pct_label_pos,
            #                       size_hint_x=None, size_hint_y=None,
            #                       halign='left', valign='center',
            #                       size=(90, 60),
            #                       text_size=(90, 60),
            #                       )
            #
            #     with pct_label.canvas.before:
            #         Rectangle(pos=pct_label_pos, size=pct_label.size, color=Color(1, 0, 0, 0.1))
            #
            #     self.add_widget(pct_label)
            #
            #     angle_start = angle_end
        else:
            pie_radius = self.size[0] / 2
            angle_start = 0

            if any([entry.value < 0 for entry in slice_data]):
                raise(ValueError('Negative values not supported for PieChart.'))

            for ix, slice_entry in enumerate(slice_data, start=0):
                # draw outer layer, negative values
                percentage = abs(slice_entry.value) / val_total
                color = slice_entry.rgba

                angle_end = angle_start + percentage * total_arc
                new_slice = PieSlice(pos=self.pos,
                                     size=(pie_radius * 2, pie_radius * 2),
                                     angle_start=angle_start, angle_end=angle_end,
                                     color=color, name=slice_entry.name,
                                     separator_width=self.separator_width,
                                     opacity=0)
                angle_start = angle_end

                self.slices.append(new_slice)
                self.add_widget(self.slices[ix])

        # redraw 0 degree slice separator
        with self.canvas.after:
            Color((1, 1, 1, 1))
            pie_center = (self.pos[0] + pie_radius, self.pos[1] + pie_radius)

            Line(points=[pie_center[0], pie_center[1],  # pie center
                         pie_center[0],  # pie perimeter x
                         pie_center[1] + pie_radius],  # pie perimeter y
                 width=self.separator_width)


class PieSlice(Widget):
    def __init__(self, pos, color, size, angle_start, angle_end, name, separator_width, **kwargs):
        """
        A widget that represents a slice of a pie chart.

        :param pos: Lower left corner of pie.
        :param color: Color (rgba) of slice.
        :param size: Size (width, height) of pie.
        :param angle_start: Starting angle of slice in degrees (0 = positive y-axis).
        :param angle_end: Ending angle of slice in degrees (0 = positive y-axis).
        :param name: Label of data corresponding to slice.
        :param separator_width: Width of line separating adjacent pie slices (px).
        :param kwargs:
        """
        super(PieSlice, self).__init__(**kwargs)

        self.name = name

        with self.canvas.before:
            # draw slice
            Color(*color)
            self.slice = Ellipse(pos=pos, size=size, angle_start=angle_start, angle_end=angle_end)

            # draw slice separator
            Color((1, 1, 1, 1))
            pie_radius = size[0]/2
            pie_center = (pos[0] + pie_radius, pos[1] + pie_radius)

            from math import cos, sin, pi

            Line(points=[pie_center[0], pie_center[1],
                         pie_center[0] + pie_radius*sin(angle_start*pi/180),
                         pie_center[1] + pie_radius*cos(angle_start*pi/180)],
                 width=separator_width)

        #self.bind(size=self._update_slice, pos=self._update_slice)

    # def _update_slice(self, instance, value):
    #     print(instance.pos, instance.size)
    #     self.slice.pos = instance.pos
    #     self.slice.size = instance.size


class ChartLegend(GridLayout):
    def __init__(self, position, key_height=30, font_size=20, **kwargs):
        super(ChartLegend, self).__init__(**kwargs)

        # Layout properties
        self.cols = 1
        self.padding = (25, self.height/3)
        self.spacing = (0, 10)

        self.pos = position
        self.key_height = key_height
        self.font_size = font_size

        #self.leg_pos = position

    def gen_legend(self, legend_data):
        # background to check widgets size and position
        # with self.canvas.before:
        #     Rectangle(pos=self.pos, size=self.size, color=Color(0.9, 0.9, 0.7, 0.3))
        #     Color(0, 0, 0, 1)
        #     Line(rectangle=(self.pos[0], self.pos[1], self.size[0], self.size[1]))

        LegEntry = namedtuple('LegEntry', ['text', 'rgba'])
        legend_entries = [LegEntry._make(x) for x in legend_data]

        for entry in legend_entries:
            color = entry.rgba
            legend_text = entry.text

            # leg_entry = LegendEntry(color=color, text=legend_text,
            #                         key_height=(self.leg_size[1]-2*self.padding[1]-self.spacing[1]*(len(legend_entries)-1))/len(legend_entries),
            #                         legend_width=self.leg_size[0])
            leg_entry = LegendEntry(color=color, text=legend_text,
                                    key_height=self.key_height,
                                    legend_width=self.size[0],
                                    font_size=self.font_size,
                                    )

            self.row_default_height = self.key_height
            self.row_force_default = True

            self.add_widget(leg_entry)


class LegendEntry(RelativeLayout):
    def __init__(self, color, text, key_height, legend_width, font_size=20, **kwargs):
        super(LegendEntry, self).__init__(**kwargs)

        self.key_color = color
        self.key_text = text
        self.key_height = key_height
        self.key_width = legend_width

        bx = GridLayout(cols=2, row_default_height=self.key_height, row_force_default=True, spacing=10)

        col_key = BoxLayout(size_hint_x=None, width=self.key_height, height=self.key_height)

        # draw key color
        with col_key.canvas:
            Rectangle(pos=col_key.pos, size=col_key.size, color=Color(*self.key_color))

        # add legend label
        col_text = Label(text=self.key_text, color=(0, 0, 0, 1),
                         pos=(self.key_height+10, 0),
                         size_hint_x=None, size_hint_y=1,
                         halign='left',
                         width=self.key_width - self.key_height,
                         height=self.key_height,
                         font_size=font_size,
                         )

        #col_text.size = col_text.texture_size
        col_text.text_size = col_text.size

        bx.add_widget(col_key)
        bx.add_widget(col_text)

        self.add_widget(bx)

        self.key_label = col_text
        self.key_square = col_key


class RateScheduleChart(Chart):
    def __init__(self, y_axis_format=None, legend_width=90, tile_spacing=2, x_padding=35, y_padding=35, **kwargs):
        """
        A rate schedule chart using colored tiles to communicate the rate tier.

        :param y_axis_format: str; format string for the y-axis labels. Namely for determining precision/separators/etc.
        :param tile_spacing: int (px) representing the space between adjacent squares.
        :param x_padding: int (px) representing the space between the left and right edges of the chart and chart elements.
        :param y_padding: int (px) representing the space between the top and bottom edges of the chart and chart elements.
        """
        super(RateScheduleChart, self).__init__(**kwargs)

        # sets default y-axis label format to two decimal place precision (fixed point)
        if not y_axis_format:
            self.y_axis_format = '{0:.2f}'
        else:
            self.y_axis_format = y_axis_format

        self.tile_spacing = tile_spacing
        self.x_padding = x_padding
        self.y_padding = y_padding
        self.legend_width = legend_width

        self.tiles = []

    def generate_tiles(self, schedule_data, category_colors, labels):
        """
        Computes the size and positions of the tiles using schedule_data. Adds the tiles and axis labels to RateScheduleChart.

        :param schedule_data: List of lists or 2D NumPy array.
        :param category_colors: List of rgba tuples to label categories.
        :param labels: List of label strings for each row.
        """
        n_rows, n_cols = schedule_data.shape        

        # Compute origin coordinates.
        x0 = self.legend_width + 3*self.x_padding
        self.max_height = self.height - self.y_padding

        # Calculate the total amount of width and height available for allocating to tiles and distribute it evenly.
        self.tile_width = (self.width - x0 - self.x_padding - self.tile_spacing*(n_cols - 1))/n_cols
        self.tile_height = (self.max_height - self.y_padding - self.tile_spacing*(n_rows - 1))/n_rows

        self.tiles = []

        # Iterate over each tile
        for iy, row in enumerate(schedule_data, start=0):
            row_label = Label(pos=(0.80*x0 - self.width/2, self.max_height/2 - (iy+0.5)*(self.tile_height + self.tile_spacing)),
            text=labels[iy], color=[0, 0, 0, 1], font_size=12, height=self.tile_height)
            self.add_widget(row_label)

            for ix, tile_value in enumerate(row, start=0):
                pos = (x0 + ix*(self.tile_width + self.tile_spacing), self.max_height - iy*(self.tile_height + self.tile_spacing))
                
                tile_widget = ScheduleTile(size=(self.tile_width, self.tile_height), pos=pos, rgba=category_colors[tile_value])
                # bar_label = Label(pos=(x0 - (self.width - self.tile_width)/2 + ix*(self.tile_width + self.tile_spacing), self.max_height/2 - iy*(self.tile_height + self.tile_spacing)), text=str(tile_value), color=[0, 0, 0, 1])

                self.tiles.append(tile_widget)
                self.add_widget(tile_widget)
                # self.add_widget(bar_label)

                if iy == 0:
                    # Column labels for hours.
                    col_label = Label(pos=(x0 - (self.width - self.tile_width)/2 + ix*(self.tile_width + self.tile_spacing), self.max_height/2 + self.y_padding/2),
                              text=str(ix).zfill(2), color=[0, 0, 0, 1], font_size=12)
                    self.add_widget(col_label)

    def draw_chart(self, schedule_data, category_colors, labels, legend_labels=None):
        """Draws the RateScheduleChart."""
        # Clear all widgets from the RateScheduleChart.
        while len(self.children) > 0:
            for widget in self.children:
                self.remove_widget(widget)

        # with self.canvas.before:
        #     Color(1, 1, 1, 1)
        #     Rectangle(size=self.size, pos=self.pos)

        self.generate_tiles(schedule_data, category_colors, labels)

        def _anim_tile(tile, *args):
            anim = Animation(opacity=1, duration=1.0, t='out_back')
            anim.start(tile)

        # Animate the tiles.
        TILE_ANIM_LENGTH = 0.005

        for ix, tile in enumerate(self.tiles, start=0):
            if isinstance(tile, ScheduleTile):
                if self._do_animation:
                    Clock.schedule_once(partial(_anim_tile, tile), ix*TILE_ANIM_LENGTH)
                else:
                    tile.opacity = 1

        # Generate legend.
        leg_pos = (0, 0)
        leg_size = (self.legend_width, self.height)

        self.legend = ChartLegend(leg_pos, size=leg_size, opacity=0, key_height=15, font_size=10)
        self.add_widget(self.legend)

        # Form the legend input [[name, rgba] for each entry.
        if not legend_labels:
            legend_data = [[str(ix), color] for ix, color in enumerate(category_colors)]
        else:
            legend_data = [[legend_labels[ix], color] for ix, color in enumerate(category_colors)]

        self.legend.gen_legend(legend_data)

        def _anim_legend(legend, *args):
            if self._do_animation:
                t_anim_legend = 0.5
            else:
                t_anim_legend = 0

            anim = Animation(opacity=1, duration=t_anim_legend, t='linear')
            anim.start(legend)

        # Animate the legend opacity
        if self._do_animation:
            Clock.schedule_once(partial(_anim_legend, self.legend), len(self.tiles)*TILE_ANIM_LENGTH)
        else:
            self.legend.opacity = 1
            
        self.is_drawn = True


class ScheduleTile(Widget):
    height_final = NumericProperty

    def __init__(self, rgba=None, **kwargs):
        """
        A widget for tiles in Rate Schedule Charts.

        :param rgba: Length-4 iterable representing R, G, B, Alpha values in [0, 1].
        :param kwargs:
        """
        super(ScheduleTile, self).__init__(**kwargs)

        self.size_hint_x = None
        self.size_hint_y = None

        self.opacity = 0

        self.color = rgba

        with self.canvas.after:
            # draws the bar as a rectangle graphic
            Color(*rgba)
            self.tile = Rectangle(pos=self.pos, size=self.size)

        self.bind(pos=self._update_tile, size=self._update_tile)

    def _update_tile(self, instance, value):
        """Updates tile graphic size and position whenever widget size or position changes."""
        self.tile.pos = instance.pos
        self.tile.size = instance.size


class RateScheduleGridLayout(GridLayout):
    pass